# coding: utf-8
"""This module provides objects and helper functions for atomic calculations."""
from __future__ import annotations

import collections
import numpy as np

from io import StringIO
from dataclasses import dataclass
from typing import Any, Union, Optional, Iterable
from monty.functools import lazy_property
from monty.string import marquee  # is_string, list_strings,
from scipy.interpolate import UnivariateSpline
try :
    from scipy.integrate import cumulative_trapezoid as cumtrapz
except ImportError:
    from scipy.integrate import cumtrapz

from abipy.data import nist_database
from abipy.tools.serialization import pmg_serialize

__version__ = "0.1"
__author__ = "Matteo Giantomassi"
__maintainer__ = "Matteo Giantomassi"

__all__ = [
    "NlkState",
    "QState",
    "AtomicConfiguration",
    "RadialFunction",
    "RadialWaveFunction",
]

char2l = {
    "s": 0,
    "p": 1,
    "d": 2,
    "f": 3,
    "g": 4,
    "h": 5,
    "i": 6,
}

l2char = {
    0: "s",
    1: "p",
    2: "d",
    3: "f",
    4: "g",
    5: "h",
    6: "i",
}

# Accept strings as keys as well.
l2char.update({str(l): l2char[l] for l in l2char})


def _asl(obj: Any) -> int:
    try:
        return char2l[obj]
    except KeyError:
        return int(obj)


def states_from_string(confstr: str) -> list[QState]:
    """
    Parse a string with an atomic configuration and build a list of `QState` instance.
    """
    states = []
    tokens = confstr.split()

    if tokens[0].startswith("["):
        noble_gas = AtomicConfiguration.neutral_from_symbol(tokens.pop(0)[1:-1])
        states = noble_gas.states

    states.extend(parse_orbtoken(t) for t in tokens)
    return states


def parse_orbtoken(orbtoken: str) -> QState:
    import re
    m = re.match(r"(\d+)([spdfghi]+)(\d+)", orbtoken.strip())
    if m:
        return QState(n=m.group(1), l=m.group(2), occ=m.group(3))

    raise ValueError("Don't know how to interpret %s" % str(orbtoken))


class NlkState(collections.namedtuple("NlkState", "n, l, k")):
    """
    Named tuple storing (n, l) or (n, l, k) if relativistic pseudos.
    k is set to None if we are in non-relativistic mode.
    """

    # The relativistic code utilizes a new main program, oncvpsp_r, rather than
    # complicate oncvpsp with lots of branching.  Many arrays have acquired an
    # additional final index, and most loops over angular momenta include inner
    # loops over this "ikap=1,2" index referring to the additional quantum number of
    # the radial Dirac equations kappa (kap) =l, -(l+1) for j=l -/+ 1/2.

    def __new__(cls, n: int, l: int, k: Optional[int] = None):
        """
        Extends super.__new__ adding type conversion and default values.
        """
        if k is not None:
            if k not in (1, 2):
                raise ValueError(f"Invalid k index: {k} for l: {l}")
            if l == 0 and k != 1:
                raise ValueError(f"When l is 0, k must be 1 while it is: {k}")

        #if n <= 0:
        #    raise ValueError(f"Invalid value for n: {n}")
        if l < 0:
            raise ValueError(f"Invalid value for l: {l}")

        return super().__new__(cls, n, l, k)

    @classmethod
    def from_nlkap(cls, n: int, l: int, kap: Union[int, None]) -> NlkState:
        k = None
        if kap is not None:
            #if(ikap==1) kap=-(ll+1)
            #if(ikap==2) kap=  ll
            k = 1
            if l != 0:
                k = {-(l + 1): 1, l:2}[kap]

        return cls(n=n, l=l, k=k)

    def __repr__(self) -> str:
        return f"n={self.n}, l={self.l}" if self.k is None else f"n={self.n}, l={self.l}, k={self.k}"

    def __str__(self) -> str:
        lc = l2char[self.l]
        if self.k is None:
            return f"{self.n}{lc}"  # e.g. 2s
        else:
            return f"{self.n}{lc}{self.ksign}"  # e.g. 2s+

    @lazy_property
    def latex(self) -> str:
        lc = l2char[self.l]
        # e.g. 2s or 2s^+
        return f"${self.n}{lc}$" if self.k is None else f"${self.n}{lc}^{self.ksign}$"

    @lazy_property
    def latex_l(self) -> str:
        lc = l2char[self.l]
        # e.g. s or s^+
        return f"${lc}$"  if self.k is None else f"${lc}^{self.ksign}$"

    @lazy_property
    def ksign(self) -> str:
        return {1: "+", 2: "-"}[self.k]

    @lazy_property
    def j(self) -> int:
        """Total angular momentum"""
        l = self.l
        if self.k is None: return l
        return l - 1/2 if self.k == l else l + 1/2

    @pmg_serialize
    def as_dict(self) -> dict:
        """Makes Nlk obey the general json interface used in pymatgen for easier serialization."""
        return {"n": self.n, "l": self.l, "k": self.k}

    def get_dict4pandas(self) -> dict:
        """Dictionary used to build pandas DataFrames."""
        return {k: v for k, v in self.as_dict().items() if not k.startswith("@")}

    def from_dict(cls, d: dict) -> NlkState:
        """
        Reconstruct object from the dictionary in MSONable format produced by as_dict.
        """
        return cls(n=d["n"], l=d["l"], k=d["k"])


class QState(collections.namedtuple("QState", "n, l, occ, eig, j, s")):
    """
    This object collects the set of quantum numbers and the physical properties
    associated to a single electron in a spherically symmetric atom.

    .. attributes:

        n: Principal quantum number.
        l: Angular momentum.
        occ: Occupancy of the atomic orbital.
        eig: Eigenvalue of the atomic orbital.
        j: J quantum number. None if spin is a good quantum number.
        s: Spin polarization. None if spin is not taken into account.
    """
    # TODO
    # Spin +1, -1 or 1,2 or 0,1?

    def __new__(cls, n: int, l: int, occ: float,
                eig: Optional[float] = None,
                j: Optional[int] = None,
                s: Optional[int] = None):
        """
        Extends super.__new__ adding type conversion and default values.
        """
        eig = float(eig) if eig is not None else eig
        j = int(j) if j is not None else j
        s = int(s) if s is not None else s
        return super().__new__(cls, int(n), _asl(l), float(occ), eig=eig, j=j, s=s)

    @property
    def has_j(self) -> bool:
        return self.j is not None

    @property
    def has_s(self) -> bool:
        return self.s is not None


class AtomicConfiguration:
    """
    Atomic configuration of an all-electron atom.
    """
    def __init__(self, Z: int, states: list[QState]) -> None:
        """
        Args:
            Z: Atomic number.
            states: List of :class:`QState` instances.
        """
        self.Z = Z
        self.states = states

    @classmethod
    def from_string(cls, Z: int, string: str,
                    has_s: bool = False, has_j: bool = False) -> AtomicConfiguration:
        if not has_s and not has_j:
            # Ex: [He] 2s2 2p3
            states = states_from_string(string)
        else:
            raise NotImplementedError("")

        return cls(Z, states)

    @classmethod
    def neutral_from_symbol(cls, symbol: Union[str, int]) -> AtomicConfiguration:
        """
        symbol: str or int
            Can be a chemical symbol (str) or an atomic number (int).
        """
        entry = nist_database.get_neutral_entry(symbol)
        states = [QState(n=s[0], l=s[1], occ=s[2]) for s in entry.states]
        return cls(entry.Z, states)

    def __str__(self) -> str:
        lines = ["%s: " % self.Z]
        lines += [str(state) for state in self]
        return "\n".join(lines)

    def __len__(self) -> int:
        return len(self.states)

    def __iter__(self) -> Iterable:
        return self.states.__iter__()

    def __eq__(self, other: AtomicConfiguration) -> bool:
        if len(self.states) != len(other.states):
            return False

        return (self.Z == other.Z and
                all(s1 == s2 for s1, s2 in zip(self.states, other.states)))

    def __ne__(self, other: AtomicConfiguration) -> bool:
        return not self == other

    def copy(self) -> AtomicConfiguration:
        """Shallow copy of self."""
        return AtomicConfiguration(self.Z, [s for s in self.states])

    @property
    def symbol(self) -> str:
        """Chemical symbol"""
        return nist_database.symbol_from_Z(self.Z)

    @property
    def spin_mode(self) -> str:
        """
        unpolarized: Spin-unpolarized calculation.
        polarized: Spin-polarized calculation.
        """
        for state in self:
            # FIXME
            if state.s is not None and state.s == 2:
                return "polarized"
        return "unpolarized"

    @property
    def echarge(self) -> float:
        """Electronic charge (float < 0 )."""
        return -sum(state.occ for state in self)

    @property
    def isneutral(self) -> bool:
        """True if self is a neutral configuration."""
        return abs(self.echarge + self.Z) < 1.e-8

    def add_state(self, **qnumbers) -> None:
        """Add a list of :class:`QState` instances to self."""
        self._push(QState(**qnumbers))

    def remove_state(self, **qnumbers) -> None:
        """Remove a quantum state from self."""
        self._pop(QState(**qnumbers))

    def _push(self, state) -> None:
        # TODO check that ordering in the input does not matter!
        if state in self.states:
            raise ValueError("state %s is already in self" % str(state))
        self.states.append(state)

    def _pop(self, state) -> None:
        try:
            self.states.remove(state)
        except ValueError:
            raise


class RadialFunction:
    """
    A RadialFunction has a name, a radial mesh and values defined on this mesh.
    """

    def __init__(self, name: str, rmesh, values):
        """
        Args:
            name: Name of the function (string).
            rmesh: Iterable with the points of the radial mesh.
            values: Iterable with the values of the function on the radial mesh.
        """
        self.name = name
        self.rmesh = np.ascontiguousarray(rmesh)
        self.values = np.ascontiguousarray(values)
        assert len(self.rmesh) == len(self.values)

    @classmethod
    def from_filename(cls, filename: str, rfunc_name=None, cols=(0, 1)) -> RadialFunction:
        """
        Initialize the object reading data from filename (txt format).

        Args:
            filename: Path to the file containing data.
            rfunc_name: Optional name for the RadialFunction (defaults to filename)
            cols: List with the index of the columns containing the radial mesh and the values.
        """
        data = np.loadtxt(filename)
        rmesh, values = data[:,cols[0]], data[:,cols[1]]
        name = filename if rfunc_name is None else rfunc_name
        return cls(name, rmesh, values)

    def __len__(self) -> int:
        return len(self.values)

    def __iter__(self) -> Iterable:
        """Iterate over (rpoint, value)."""
        return iter(zip(self.rmesh, self.values))

    def __getitem__(self, rslice):
        return self.rmesh[rslice], self.values[rslice]

    def __repr__(self) -> str:
        return "<%s, name=%s at %s>" % (self.__class__.__name__, self.name, id(self))

    def __str__(self) -> str:
        stream = StringIO()
        self.pprint(stream=stream)
        return stream.getvalue()

    #def __add__(self, other):
    #def __sub__(self, other):
    #def __mul__(self, other):

    def __abs__(self) -> RadialFunction:
        return self.__class__(self.rmesh, np.abs(self.values))

    @property
    def to_dict(self) -> dict:
        return dict(
            name=str(self.name),
            rmesh=list(self.rmesh),
            values=list(self.values),
        )

    def pprint(self, what: str = "rmesh+values", stream=None) -> None:
        """pprint method (useful for debugging)"""
        from pprint import pprint
        if "rmesh" in what:
            pprint("rmesh:", stream=stream)
            pprint(self.rmesh, stream=stream)

        if "values" in what:
            pprint("values:", stream=stream)
            pprint(self.values, stream=stream)

    @property
    def rmax(self) -> float:
        """Outermost point of the radial mesh."""
        return self.rmesh[-1]

    @property
    def rsize(self) -> float:
        """Size of the radial mesh."""
        return len(self.rmesh)

    @property
    def minmax_ridx(self) -> tuple[int, int]:
        """
        Returns the indices of the values in a list with the maximum and minimum value.
        """
        minimum = min(enumerate(self.values), key=lambda s: s[1])
        maximum = max(enumerate(self.values), key=lambda s: s[1])
        return minimum[0], maximum[0]

    @property
    def inodes(self) -> list[int]:
        """"
        List with the index of the nodes of the radial function.
        """
        inodes = []
        for i in range(len(self.values)-1):
            if self.values[i] * self.values[i+1] <= 0:
                inodes.append(i)
        return inodes

    @lazy_property
    def spline(self):
        """Cubic spline."""
        #return UnivariateSpline(self.rmesh, self.values, s=0)
        return UnivariateSpline(self.rmesh, self.values, s=None)

    @lazy_property
    def roots(self):
        """Return the zeros of the spline."""
        return self.spline.roots()

    def get_peaks(self, **kwargs):
        """
        """
        from scipy.signal import find_peaks
        inds, properties = find_peaks(self.values, **kwargs)
        xs = self.rmesh[inds] if len(inds) else []
        ys = self.values[inds] if len(inds) else []
        return Peaks(xs=xs, ys=ys, inds=inds, properties=properties)

    def derivatives(self, r):
        """Return all derivatives of the spline at the point r."""
        return self.spline.derivatives(r)

    def integral(self) -> RadialFunction:
        r"""
        Cumulatively integrate y(x) using the composite trapezoidal rule.

        Returns:
            `Function1d` with :math:`\int y(x) dx`
        """
        integ = cumtrapz(self.values, x=self.rmesh)
        pad_intg = np.zeros(len(self.values))
        pad_intg[1:] = integ

        return self.__class__(self.rmesh, pad_intg)

    def integral3d(self, a=None, b=None):
        """
        Return definite integral of the spline of (r**2 values**2) between two given points a and b

        Args:
            a: First point. rmesh[0] if a is None
            b: Last point. rmesh[-1] if a is None
        """
        a = self.rmesh[0] if a is None else a
        b = self.rmesh[-1] if b is None else b
        r2v2_spline = UnivariateSpline(self.rmesh, (self.rmesh * self.values) ** 2, s=0)

        return r2v2_spline.integral(a, b)

    def ifromr(self, rpoint) -> int:
        """
        The index of the point in the radial mesh.
        """
        for (i, r) in enumerate(self.rmesh):
            if r > rpoint:
                return i - 1

        if rpoint == self.rmesh[-1]:
            return len(self.rmesh)
        else:
            raise ValueError("Cannot find %s in rmesh" % rpoint)

    def ir_small(self, abs_tol: float = 0.01) -> int:
        """
        Returns the rightmost index where the abs value of the wf becomes greater than abs_tol

        Args:
            abs_tol: Absolute tolerance.

        .. warning::

            Assumes that self.values are tending to zero for r --> infinity.
        """
        for i in range(len(self.rmesh)-1, -1, -1):
            if abs(self.values[i]) > abs_tol:
                break
        return i

    def r2f2_integral(self):
        """
        Cumulatively integrate r**2 f**2(r) using the composite trapezoidal rule.
        """
        integ = cumtrapz(self.rmesh**2 * self.values**2, x=self.rmesh)
        pad_intg = np.zeros(len(self))
        pad_intg[1:] = integ

        return pad_intg

    def r2f_integral(self):
        """
        Cumulatively integrate r**2 f(r) using the composite trapezoidal rule.
        """
        integ = cumtrapz(self.rmesh**2 * self.values, x=self.rmesh)
        pad_intg = np.empty(len(self))
        pad_intg[1:] = integ
        return pad_intg

    def get_intr2j0(self, ecut: float, numq: float = 3001):
        r"""
        Compute 4\pi\int[(\frac{\sin(2\pi q r)}{2\pi q r})(r^2 n(r))dr].
        """
        qmax = np.sqrt(ecut / 2) / np.pi
        qmesh = np.linspace(0, qmax, num=numq, endpoint=True)
        outs = np.empty(len(qmesh))

        # Treat q == 0. Note that rmesh[0] > 0
        f = 4 * np.pi * self.rmesh**2 * self.values
        outs[0] = cumtrapz(f, x=self.rmesh)[-1]

        for i, q in enumerate(qmesh[1:]):
            twopiqr = 2 * np.pi * q * self.rmesh
            f = 4 * np.pi * (np.sin(twopiqr) / twopiqr) * self.rmesh**2 * self.values
            outs[i+1] = cumtrapz(f, x=self.rmesh)[-1]

        from abipy.core.func1d import Function1D
        ecuts = 2 * np.pi**2 * qmesh**2
        return Function1D(ecuts, outs)


class RadialWaveFunction(RadialFunction):
    """
    Extends :class:`RadialFunction` adding info on the set of quantum numbers.
    and methods specialized for electronic wavefunctions.
    """

    def __init__(self, nlk: NlkState, name: str, rmesh, values):
        super().__init__(name, rmesh, values)
        self.nlk = nlk


@dataclass
class Peaks:
    """
    Store information on the peaks of the radial functions.
    """
    xs: np.ndarray     # Absissas of the peaks
    ys: np.ndarray     # Values of the peaks
    inds: np.ndarray   # Indices of the peaks.
    properties: dict   # Dict with peaks properties.

    def __bool__(self) -> bool:
        return bool(len(self.xs))

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, title=None, verbose=0) -> str:
        """
        String representation.
        """
        lines = []; app = lines.append
        if title is not None: app(marquee(title, mark="="))

        if self:
            app(f"last peak at: {round(self.xs[-1], 2)}, num peaks: {len(self.xs)}")

        return "\n".join(lines)




# coding: utf-8
"""
Objects to plot electronic, vibrational and e-ph properties.
"""
from __future__ import annotations

import numpy as np

from abipy.tools.plotting import (add_fig_kwargs, get_axarray_fig_plt, set_axlims, set_visible, ax_share)
from abipy.tools import duck
from abipy.electrons.ebands import ElectronBands
from abipy.dfpt.ddb import DdbFile
from abipy.dfpt.phonons import PhbstFile, PhdosFile
from abipy.eph.sigeph import SigEPhFile
from abipy.tools.typing import Figure


class EphPlotter:
    """
    This object provides methods to plot electron and phonons for a single system.
    An EphPlotter has:

        - An |ElectronBands| on a k-path
        - An |ElectronBands| on a k-mesh (optional)
        - A |PhbstFile| with phonons along a q-path.
        - A |PhdosFile| with different kinds of phonon DOSes.

    EphPlotter uses these objects/files and other inputs/files provided by
    the user to generate matplotlib plots related to e-ph interaction.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: EphPlotter
    """
    @classmethod
    def from_ddb(cls, ddb, ebands_kpath, ebands_kmesh=None, **kwargs) -> EphPlotter:
        """
        Build the object from the ddb file, invoke anaddb to get phonon properties.
        This entry point is needed to have phonon plots with LO-TO splitting
        as AbiPy will generate an anaddb input with the different q --> 0 directions
        required in phbands.plot to plot the LO-TO splitting correctly.

        Args:
            ddb: |DdbFile| or filepath.
            ebands_kpath: |ElectronBands| with energies on a k-path or filepath.
            ebands_kpath: (optional) |ElectronBands| with energies on a k-mesh or filepath.
            kwargs: Passed to anaget_phbst_and_phdos_files
        """
        ddb = DdbFile.as_ddb(ddb)
        phbst_file, phdos_file = ddb.anaget_phbst_and_phdos_files(**kwargs)
        return cls(ebands_kpath, phbst_file, phdos_file, ebands_kmesh=ebands_kmesh)

    def __init__(self, ebands_kpath, phbst_file, phdos_file, ebands_kmesh=None):
        self.eb_kpath = ElectronBands.as_ebands(ebands_kpath)
        self.eb_kmesh = ElectronBands.as_ebands(ebands_kmesh) if ebands_kmesh is not None else None

        self.phbst_file = phbst_file
        if duck.is_string(self.phbst_file):
            self.phbst_file = PhbstFile(self.phbst_file)
        self.phb_qpath = self.phbst_file.phbands

        self.phdos_file = phdos_file
        if duck.is_string(self.phdos_file):
            self.phdos_file = PhdosFile(phdos_file)

    @add_fig_kwargs
    def plot(self, eb_ylims=None, **kwargs) -> Figure:
        """
        Plot electrons with possible (phonon-mediated) scattering channels for the CBM and VBM.
        Also plot phonon band structure and phonon PJDOS.

        Args:
            eb_ylims: Set the data limits for the y-axis of the electron band. Accept tuple e.g. ``(left, right)``
                or scalar e.g. ``left``. If None, limits are selected automatically.

        Return: |matplotlib-Figure|
        """
        # Build grid. Share y-axis for Phbands and Phdos
        import matplotlib.pyplot as plt
        fig = plt.figure()
        ax0 = plt.subplot2grid((3, 3), (0, 0), colspan=3, rowspan=2)
        ax1 = plt.subplot2grid((3, 3), (2, 0), colspan=2, rowspan=1)
        ax2 = plt.subplot2grid((3, 3), (2, 2), colspan=1, rowspan=1)
        ax1.get_shared_y_axes().join(ax1, ax2)

        # Plot electrons with possible e-ph scattering channels.
        self.eb_kpath.plot(ax=ax0, with_gaps=True, ylims=eb_ylims, max_phfreq=self.phb_qpath.maxfreq, show=False)

        # Plot phonon bands
        self.phb_qpath.plot(ax=ax1, show=False)
        #ax1.yaxis.set_visible(False)
        #set_visible(ax1, False, "ylabel")

        # Plot phonon PJDOS
        self.phdos_file.plot_pjdos_type(ax=ax2, fontsize=8, exchange_xy=True, show=False)
        set_visible(ax2, False, "ylabel")
        ax2.tick_params("y", left=False, labelleft=False)
        ax2.tick_params("y", right=True, labelright=True)

        # Adjust y-limits for phonons
        ylims = self.phb_qpath.minfreq, self.phb_qpath.maxfreq + 0.1 * abs(self.phb_qpath.maxfreq)
        for ax in (ax1, ax2):
            set_axlims(ax, ylims, "y")

        return fig

    @add_fig_kwargs
    def plot_phonons_occ(self, temps=(100, 200, 300, 400), **kwargs) -> Figure:
        """
        Plot phonon band structure with markers proportional to the occupation
        of each phonon mode for different temperatures.

        Args:
            temps: List of temperatures in Kelvin.

        Return: |matplotlib-Figure|
        """
        temps = np.array(temps)
        ntemp = len(temps)

        # Build plot grid.
        num_plots, ncols, nrows = ntemp, 1, 1
        if num_plots > 1:
            ncols = 2
            nrows = (num_plots // ncols) + (num_plots % ncols)

        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                                sharex=True, sharey=True, squeeze=False)
        ax_list = ax_list.ravel()

        for ax, temp in zip(ax_list, temps.ravel()):
            self.phb_qpath.plot(ax=ax, units="eV", temp=temp, fontsize=8, show=False)

        return fig

    @add_fig_kwargs
    def plot_linewidths_sigeph(self, sigeph, eb_ylims=None, **kwargs) -> Figure:
        """
        Plot e-bands + e-DOS + Im(Sigma_{eph}) + phonons + gkq^2

        Args:
            sigeph: |SigephFile| or string with path to file.
            eb_ylims: Set the data limits for the y-axis of the electron band. Accept tuple e.g. ``(left, right)``
                or scalar e.g. ``left``. If None, limits are selected automatically.
        """
        closeit = False
        if duck.is_string(sigeph):
            sigeph = SigEPhFile.from_file(sigeph)
            closeit = True

        # Build grid. share y-axis for Phbands and Phdos
        import matplotlib.pyplot as plt
        fig = plt.figure()

        # Electrons
        ax0 = plt.subplot2grid((2, 4), (0, 0), colspan=2, rowspan=1)
        ax1 = plt.subplot2grid((2, 4), (0, 2), colspan=1, rowspan=1)
        ax2 = plt.subplot2grid((2, 4), (0, 3), colspan=1, rowspan=1)
        # Share y-axis
        ax_share("y", ax0, ax1, ax2)

        # Phonons
        ax3 = plt.subplot2grid((2, 4), (1, 0), colspan=2, rowspan=1)
        ax4 = plt.subplot2grid((2, 4), (1, 2), colspan=1, rowspan=1)
        ax5 = plt.subplot2grid((2, 4), (1, 3), colspan=1, rowspan=1)
        # Share y-axis
        ax_share("y", ax3, ax4, ax5)

        e0 = "fermie"

        # Plot electrons with possible e-ph scattering channels.
        self.eb_kpath.plot(ax=ax0, e0=e0, with_gaps=True, ylims=eb_ylims, max_phfreq=self.phb_qpath.maxfreq, show=False)
        sigeph.plot_lws_vs_e0(ax=ax1, e0=e0, exchange_xy=True, show=False)
        sigeph.edos.plot(ax=ax2, e0=e0, exchange_xy=True, show=False)

        # Plot phonon bands
        self.phb_qpath.plot(ax=ax3, show=False)
        sigeph.plot_a2fw_skb_sum(ax=ax4, what="gkq2", exchange_xy=True, fontsize=8, show=False)
        # Plot phonon PJDOS
        self.phdos_file.plot_pjdos_type(ax=ax5, fontsize=8, exchange_xy=True, show=False)
        #set_visible(ax4, False, "ylabel")
        #ax4.tick_params("y", left=False, labelleft=False)
        #ax4.tick_params("y", right=True, labelright=True)

        if closeit: sigeph.close()

        return fig

    #def close(self):
    #    self.phbst_file.close()
    #    self.phdos_file.close()


#class EphMultiPlotter(object):



#!/usr/bin/env python
r"""
Plot the LRUJ results
=====================

This example shows how to parse the output file produced by lruj and plot the results


See also https://docs.abinit.org/tutorial/lruj/#43-execution-of-the-lruj-post-processinng-utility
"""
from abipy.electrons.lruj import LrujAnalyzer, LrujResults
import abipy.data as abidata
import os

# Initialize LrujResults from the main output file of lruj
outfile = abidata.ref_file("lruj_data/lruj.out")
lr = LrujResults.from_file(outfile)

#%%
# Plot the fits.

lr.plot(degrees="all", insetdegree=4, ptcolor0='blue', ptitle="Hello World", fontsize=9)

#filepaths = [
#    "tlruj_2.o_DS1_LRUJ.nc",
#    "tlruj_2.o_DS2_LRUJ.nc",
#    "tlruj_2.o_DS3_LRUJ.nc",
#    "tlruj_2.o_DS4_LRUJ.nc",
#]
#
#root = "tutorial_tlruj_1-tlruj_2-tlruj_3/"
#filepaths = [os.path.join(root, p) for p in filepaths]
#
#lruj = LrujAnalyzer(verbose=1)
#lruj.add_ncpaths("foo", filepaths)
#lruj.add_ncpaths("bar", filepaths)

#print(lruj)
#lruj.plot()



"""
Parser for the output files produced by the CP code.

See: https://www.quantum-espresso.org/Doc/INPUT_CP.html

      prefix.pos : atomic positions
      prefix.vel : atomic velocities
      prefix.for : atomic forces
      prefix.cel : cell parameters
      prefix.str : stress tensors
      prefix.evp : energies
      prefix.hrs : Hirshfeld effective volumes (ts-vdw)
      prefix.eig : eigen values
      prefix.nos : Nose-Hoover variables
      prefix.spr : spread of Wannier orbitals
      prefix.wfc : center of Wannier orbitals
      prefix.ncg : number of Poisson CG steps (PBE0)
      prefix_ndw.save/ : write restart folder
      prefix_ndr.save/ : read restart folder
"""
from __future__ import annotations

import numpy as np
import pandas as pd
import dataclasses
import abipy.core.abinit_units as abu

from pathlib import Path
from monty.functools import lazy_property
#from monty.bisect import find_le
from abipy.core.mixins import TextFile, NotebookWriter
from abipy.tools.typing import PathLike, Figure
from abipy.tools.plotting import (set_axlims, add_fig_kwargs, get_ax_fig_plt, get_axarray_fig_plt,
    get_ax3d_fig_plt, rotate_ticklabels, set_visible, plot_unit_cell, set_ax_xylabels, get_figs_plotly,
    get_fig_plotly, add_plotly_fig_kwargs, PlotlyRowColDesc, plotly_klabels, plotly_set_lims)


def parse_file_with_blocks(filepath: PathLike, block_len: int) -> tuple[int, np.ndarray, list]:
    """
    Parse a QE file whose format is:

          20    0.00241888
         0.31882936800000E+01     0.14832370390000E+02     0.12288296100000E+01
         0.78323146900000E+01     0.67870403900000E+01     0.12288296100000E+01
         0.20744346700000E+01     0.59953799200000E+01     0.47375825000000E+01
          20    0.00241888
         0.31889894893205E+01     0.14833320999242E+02     0.12271083047604E+01
         0.78345550025202E+01     0.67881458009071E+01     0.12273445304341E+01
         0.20762325213327E+01     0.59955571558384E+01     0.47335647385293E+01

    i.e. a header followed by block_len lines

    Args:
        filepath: Filename
        block_len: Number of lines in block.

    Return: (number_of_steps, array_to_be_reshaped, list_of_headers)
    """
    nsteps, data, headers = 0, [], []
    with open(filepath, "rt") as fh:
        for il, line in enumerate(fh):
            toks = line.split()
            if il % (block_len + 1) == 0:
                #assert len(toks) == 2
                headers.append([int(toks[0]), float(toks[1])])
                nsteps += 1
            else:
                data.append([float(t) for t in toks])

    return nsteps, np.array(data, dtype=float), headers


def parse_file_with_header(filepath: PathLike) -> pd.DataFrame:
    """
    Parse a file with an header followed by csv-like columns e.g. an .evp file
    """
    from io import StringIO
    sbuf = StringIO()
    with open(filepath, "rt") as fh:
        for il, line in enumerate(fh):
            if il == 0:
                if not line.startswith("#"):
                    raise ValueError(f"Line number {il=} in {filepath=} should start with '#', but got {line=}")
                names = line[1:].split()
                continue

            if line.startswith("#"):
                # Found other header, likely due to restart.
                # Check that column names agree with the first header.
                new_names = line[1:].split()
                if new_names != names:
                    raise ValueError("Line number {il=} in {filepath=} has different names\n: {new_names=}\n{names=}")
                continue

            sbuf.write(line)

    sbuf.seek(0)
    df = pd.read_csv(sbuf, delim_whitespace=True, names=names)
    sbuf.close()
    return df


@dataclasses.dataclass
class Key:
    name: str
    info: str = "No info available"
    color: str = "b"


class EvpFile(TextFile, NotebookWriter):
    """
    The evp file is a file that contains the electronic and ionic energies and pressures
    for each time step of a CP simulation.
    The evp file has the following format:

    # nfi tps(ps) ekinc Tcell(K) Tion(K) etot enthal econs
          0            0          300   -2099.0164    7.4066063   -2091.6098    19363.353    2188.2406    5.0978969
          10         0.01    557.43004   -2105.3429    13.762216   -2091.5807    16917.626    2188.2406    5.0978969
    ...

    NB: Energies are in Hartree and not in Ry.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: EvpFile
    """

    COLS_DICT = {k.name: k for k in [
         Key(name="nfi"    , color="b", info="number of iterations"),
         Key(name="ekinc"  , color="b", info="electron fake kinetic energy"),
         Key(name="temphc" , color="b", info="temperature due to “cell” kinetic energy in K"),
         Key(name="tempp"  , color="b", info="temperature due to ionic displacements within the cell"),
         Key(name="etot"   , color="b", info="dft total energy (without ionic and cell kinetic energy)"),
         Key(name="enthal" , color="b", info="etot+external_pressure*volume"),
         Key(name="econs"  , color="b", info="etot+kinetic_energy_due_to_ions_moving"),
         Key(name="econt"  , color="b", info="econs+ekinc+(thermostat_total_energy)"),
         Key(name="time(ps)", color="b", info="time"),
         Key(name="tps(ps)", color="b", info="time"),
         #Volume        Pressure(GPa)        EXX               EVDW
    ]}

    @lazy_property
    def time_key(self) -> str:
        key1 = "time(ps)"
        key2 = "tps(ps)"
        if key1 in self.df.keys(): return key1
        if key2 in self.df.keys(): return key2
        raise ValueError(f"Cannot find time key in {self.df.keys()}")

    @lazy_property
    def df(self) -> pd.DataFrame:
        """Dataframe with the results."""
        # TODO: Handle restart
        df = parse_file_with_header(self.filepath)
        return df

    @lazy_property
    def times(self) -> np.ndarray:
        """Array with times in ps units."""
        return np.array(self.df["tps(ps)"].values, dtype=float)

    def to_string(self, verbose=0) -> str:
        """String representation with verbosity level verbose."""
        lines = []; app = lines.append
        app(self.df.describe(percentiles=None).to_string())
        if verbose:
            for name, col in self.COLS_DICT.items():
                app(f"{name}: {col.info}")
        else:
            app("Use verbose to print the meaning of each column")
        return "\n".join(lines)

    @add_fig_kwargs
    def plot(self, **kwargs) -> Figure:
        """
        """
        ax_mat = self.df.plot.line(x=self.time_key, subplots=True,
                                   y=[k for k in self.df.keys() if k not in ("nfi", self.time_key)])
        return ax_mat.ravel()[0].get_figure()

        """
        # plot(self, ax_mat=None, t0=0.0, **kwargs) -> Figure:
        times = self.df["tps(ps)"]
        it0 = find_le(times, t0) if t0 != 0.0 else 0
        nrows, ncols = (1, 1)
        ax_mat, fig, plt = get_axarray_fig_plt(ax_array=ax_mat, nrows=nrows, ncols=ncols,
                                               sharex=True, sharey=False, squeeze=False)
        ax = ax_mat[0,0]

        yvals = self.df["ekinc"]
        # Plot ionic temperature.
        ax.plot(times[it0:], yvals[it0:], label="$Temp_{ions}$", c='tab:red')
        ax.set_xlabel('time (ps)')
        ax.set_ylabel('temperature (K)')
        material = "foobar"
        ax.text(0.35, 0.9, material,
                verticalalignment='bottom', horizontalalignment='right',
                transform=ax.transAxes, color='black', fontsize=15)
        ax.legend(loc=1, bbox_to_anchor=(1, 1))
        #for irow in range(nrows):
        #    for iax, (ax, data, label) in enumerate(zip(ax_mat[irow], select_irow[irow], label_irow[irow])):
        return fig
        """

    @add_fig_kwargs
    def plot_hist(self, **kwargs) -> Figure:
        """Plot one histogram per column."""
        ax_mat = self.df.hist(column=[k for k in self.df.keys() if k not in ("nfi", self.time_key)])
        return ax_mat.ravel()[0].get_figure()

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        yield self.plot(show=False)
        yield self.plot_hist(show=False)

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            nbv.new_code_cell("evp = abilab.abiopen('%s')" % self.filepath),
            nbv.new_code_cell("evp.plot();"),
            nbv.new_code_cell("evp.plot_hist();"),
        ])

        return self._write_nb_nbpath(nb, nbpath)


def traj_to_qepos(traj_filepath: PathLike, pos_filepath: PathLike) -> None:
    """
    Convert an ASE trajectory file to QE POS file.

    Args:
        traj: Name of ASE trajectory file or Trajectory instance.
        pos_filepath: Name of output file.
    """
    from ase.io import read
    traj = read(traj_filepath, index=":")

    nsteps, natoms = len(traj), len(traj[0])
    pos_tac = np.empty((nsteps, natoms, 3))
    for it, atoms in enumerate(traj):
        pos_tac[it] = atoms.positions

    with open(str(pos_filepath), "wt") as fh:
        for it in range(nsteps):
            fh.write(str(it) + '\n')
            for ia in range(natoms):
                fh.write(str(pos_tac[it,ia,0]) + ' ' + str(pos_tac[it,ia,1]) + ' ' + str(pos_tac[it,ia,2]) + '\n')


class Qe2Extxyz:
    """
    Convert QE/CP output files into ASE extended xyz format.

    Example:

        from abipy.dynamics.cpx import Qe2Extxyz
        converter = Qe2Extxyz.from_input("cp.in", code="cp")
        converter.write_xyz("extended.xyz", take_every=1)
    """

    @classmethod
    def from_input(cls, qe_input, code, prefix=None):
        """
        Build object from QE/CP input assuming all the other output files
        are located in the same directory with the given prefix.
        """
        qe_input = Path(str(qe_input)).absolute()
        directory = qe_input.cwd()
        from ase.io.espresso import read_fortran_namelist
        with open(qe_input, "rt") as fh:
            sections, card_lines = read_fortran_namelist(fh)

        # https://www.quantum-espresso.org/Doc/INPUT_CP.html
        control = sections["control"]
        calculation = control["calculation"]
        #outdir = control["outdir"]

        default_prefix = "cp" if code == "cp" else "pwscf"
        if prefix is None:
            prefix = control.get("prefix", default_prefix)

        pos_filepath = directory / f"{prefix}.pos"
        for_filepath = directory / f"{prefix}.for"
        cell_filepath = directory / f"{prefix}.cel"
        evp_filepath = directory / f"{prefix}.evp"
        str_filepath = directory / f"{prefix}.str"
        if not str_filepath.exists():
            # File with stresses is optional.
            str_filepath = None

        return cls(code, qe_input, pos_filepath, for_filepath, cell_filepath, evp_filepath, str_filepath=str_filepath)

    def __init__(self, code, qe_input, pos_filepath, for_filepath, cell_filepath, evp_filepath, str_filepath=None):
        """
        Args:
            code: Either "pwscf" or "cp".
            qe_input: QE/CP input file
            pos_filepath: File with cartesian positions.
            for_filepath: File with cartesian forces.
            cell_filepath: File with lattice vectors.
            evp_filepath: File with energies
        """
        print(f"""
Reading symbols from: {qe_input=}
Reading positions from: {pos_filepath=}
Reading forces from: {for_filepath=}
Reading cell from: {cell_filepath=}
Reading energies from: {evp_filepath=}
Reading stresses from: {str_filepath=}
        """)

        # Parse input file to get initial_atoms and symbols.
        from ase.io.espresso import read_espresso_in
        with open(qe_input, "rt") as fh:
            self.initial_atoms = read_espresso_in(fh)
            natom = len(self.initial_atoms)
            #print("initial_atoms:", self.initial_atoms)

        # NB: in QE/CP lengths are in Bohr
        # but CP uses Hartree for energies whereas PW uses Rydberg.
        e_fact = 1.0 if code == "cp" else 0.5

        # Parse lattice vectors NB: in the cel file, lattice vectors are stored as column-vectors.
        # so he have to transpose the data as ASE uses row-vectors.
        cell_nsteps, cells, cell_headers = parse_file_with_blocks(cell_filepath, 3)
        cells = np.reshape(cells, (cell_nsteps, 3, 3)) * abu.Bohr_Ang
        self.cells = np.transpose(cells, (0, 2, 1)).copy()

        # Get energies from the evp file and convert from Ha to eV.
        with EvpFile(evp_filepath) as evp:
            self.energies = evp.df["etot"].values * (e_fact * abu.Ha_to_eV)
            if len(self.energies) != cell_nsteps:
                raise RuntimeError(f"{len(self.energies)=} != {cell_nsteps=}")

        # Parse Cartesian positions (Bohr --> Ang).
        pos_nsteps, pos_cart, pos_headers = parse_file_with_blocks(pos_filepath, natom)
        self.pos_cart = np.reshape(pos_cart, (pos_nsteps, natom, 3)) * abu.Bohr_Ang
        if pos_nsteps != cell_nsteps:
            raise RuntimeError(f"{pos_nsteps=} != {cell_nsteps=}")

        # Parse Cartesian forces (Ha/Bohr --> eV/Ang).
        for_nsteps, forces_list, for_headers = parse_file_with_blocks(for_filepath, natom)
        self.forces_step = np.reshape(forces_list, (for_nsteps, natom, 3)) * (e_fact * abu.Ha_to_eV / abu.Bohr_Ang)
        if for_nsteps != cell_nsteps:
            raise RuntimeError(f"{for_nsteps=} != {cell_nsteps=}")

        # Optionally, parse stress. In ASE, stress is eV/Ang^3.
        self.stresses = None
        if str_filepath is not None:
            str_nsteps, stresses, str_headers = parse_file_with_blocks(str_filepath, 3)
            self.stresses = np.reshape(stresses, (str_nsteps, 3, 3)) * (e_fact * abu.Ha_to_eV / abu.Bohr_Ang**3)
            if str_nsteps != cell_nsteps:
                raise RuntimeError(f"{str_nsteps=} != {cell_nsteps=}")

        self.nsteps = cell_nsteps

    def write_xyz(self, xyz_filepath, take_every=1, pbc=True) -> None:
        """
        Write results in ASE extended xyz format.

        Args:
            xyz_filepath: Name of the XYZ file.
            take_every: Used to downsample the trajectory.
            pbc: Values of pbc.
        """
        print(f"Writing results in extended xyz format in: {xyz_filepath}")
        from ase.io import write
        with open(xyz_filepath, "wt") as fh:
            for istep, atoms in enumerate(self.yield_atoms(pbc)):
                if istep % take_every != 0: continue
                write(fh, atoms, format='extxyz', append=True)

    def yield_atoms(self, pbc):
        """Yields ASE atoms along the trajectory."""
        from ase.atoms import Atoms
        from ase.calculators.singlepoint import SinglePointCalculator
        for istep in range(self.nsteps):
            atoms = Atoms(symbols=self.initial_atoms.symbols,
                          positions=self.pos_cart[istep],
                          cell=self.cells[istep],
                          pbc=pbc,
                          )

            # Attach calculator with results.
            atoms.calc = SinglePointCalculator(atoms,
                                               energy=self.energies[istep],
                                               free_energy=self.energies[istep],
                                               forces=self.forces_step[istep],
                                               stress=self.stresses[istep] if self.stresses is not None else None,
                                               )
            yield atoms


def downsample_xyz(input_xyz, take_every, output_xyz, skip_head=None, verbose=1) -> int:
    """
    Downsample an XYZ file. Return the number of configurations written in the new file.

    Args:
        input_xyz: Input XYZ file.
        take_every: Extract configurations from input_xyz every `take_every` step.
        output_xyz: Output XYZ file.
        skip_head: If not None, skip the first skip_head configurations.
        verbose: Verbosity level.
    """
    from ase.io import write, iread
    count = 0
    with open(output_xyz, "wt") as out_xyz:
        for istep, atoms in enumerate(iread(input_xyz)):
            if istep % take_every != 0: continue
            if skip_head is not None and (istep + 1) <= skip_head: continue
            count += 1
            write(out_xyz, atoms, format='extxyz', append=True)

    if verbose: print(f"Wrote {count=} configurations to {output_xyz=} with {take_every=} and {skip_head=}")
    return count




""""Panels for phonon-related objects."""
from __future__ import annotations

import param
import panel as pn
import panel.widgets as pnw

from abipy.panels.core import AbipyParameterized, mpl, ply, dfc, depends_on_btn_click


class PhononBandsPlotterPanel(AbipyParameterized):
    """
    """

    def __init__(self, plotter, **params):

        self.phbands_plotter_mode = pnw.Select(name="Plot Mode", value="gridplot",
                                               options=["gridplot", "combiplot", "boxplot", "combiboxplot"]) # "animate",
        self.phbands_plotter_units = pnw.Select(name="Units", value="eV",
                                                options=["eV", "meV", "Ha", "cm-1", "Thz"])
        self.phbands_plotter_btn = pnw.Button(name="Plot", button_type='primary')

        self.plotter = plotter
        super().__init__(**params)

    @depends_on_btn_click('phbands_blotter_btn')
    def on_phbands_plot_btn(self) -> pn.Row:
        plot_mode = self.phbands_plotter_mode.value
        plotfunc = getattr(self.plotter, plot_mode, None)
        if plotfunc is None:
            raise ValueError("Don't know how to handle plot_mode: %s" % plot_mode)

        fig = plotfunc(units=self.phbands_plotter_units.value, **self.mpl_kwargs)
        df = self.plotter.get_phbands_frame(with_spglib=True)
        return pn.Row(pn.Column(mpl(fig), dfc(df)), sizing_mode='scale_width')

    def get_panel(self, as_dict=False, **kwargs):
        """Return tabs with widgets to interact with the |PhononBandsPlotter|."""
        d = {}

        ws = pn.Column(self.phbands_plotter_mode, self.phbands_plotter_units, self.phbands_plotter_btn)
        d["PhbandsPlotter"] = pn.Row(ws, self.on_phbands_plot_btn, sizing_mode='scale_width')

        if as_dict: return d

        return self.get_template_from_tabs(d, template=kwargs.get("template", None))



#!/usr/bin/env python
r"""
RPA flow with convergence studies
=================================

This script computes the RPA energy with the GWR code
and the quartic scaling algorithm (Adler-Wiser exact expression in frequency domain)
so that one can then compare the results.
"""

import os
import sys
import abipy.data as data
import abipy.abilab as abilab

from abipy import flowtk


def build_flow(options):
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_","flow_")

    from abipy.data.gwr_structures import get_gwr_structure
    symbol = "Si"
    #symbol = "LiF"
    #symbol = "Si"
    #symbol = "C"
    #symbol = "BN"
    #symbol = "MgO"
    #symbol = "GaAs"
    #symbol = "ZnO"
    structure = get_gwr_structure(symbol)

    from abipy.flowtk.psrepos import get_repo_from_name
    #pseudos = get_repo_from_name("ONCVPSP-PBE-SR-PDv0.4").get_pseudos("stringent")
    pseudos = get_repo_from_name("ONCVPSP-PBE-SR-PDv0.4").get_pseudos("standard")

    scf_input = abilab.AbinitInput(structure=structure, pseudos=pseudos)
    scf_input.set_cutoffs_for_accuracy("normal")
    ecut = scf_input["ecut"]
    scf_input.set_scf_nband_semicond()

    # Global variables.
    scf_input.set_vars(
        tolvrs=1e-8,
        paral_kgb=0,
        npfft=1,
        timopt=-1,
    )
    scf_input.set_kmesh(
        ngkpt=[1, 1, 1],
        #ngkpt=[4, 4, 4],
        #ngkpt=[8, 8, 8],
        shiftk=[0.0, 0.0, 0.0], # IMPORTANT: k-grid for GWR must be Gamma-centered.
    )

    # Get max number of PWs.
    dims, _ = scf_input.abiget_dims_spginfo()
    mpw = dims["mpw"]

    flow = flowtk.Flow(workdir=options.workdir)
    small_manager = options.manager.new_with_fixed_mpi_omp(4, 1)

    # GS-SCF run to get the DEN, followed by direct diago to obtain green_nband bands.
    from abipy.flowtk.gwr_works import DirectDiagoWork, GWRRPAConvWork
    green_nband = -1  # -1 this means full diago
    diago_work = DirectDiagoWork.from_scf_input(scf_input, green_nband)
    wfk_node = diago_work[1]
    diago_work[0].set_manager(small_manager)
    flow.register_work(diago_work)

    gwr_ntau_list = list(range(6, 34, 2))
    gwr_ntau_list = [6, 8]
    nband = int(mpw*0.8)
    ecuteps = 12

    # GWR with different values of regterm
    for gwr_regterm in (0.0, 1e-6):
        work = GWRRPAConvWork.from_scf_input_ntaus(scf_input, gwr_ntau_list,
                                                   nband=nband, ecuteps=ecuteps,
                                                   den_node=diago_work[0], wfk_node=wfk_node,
                                                   gwr_kwargs=dict(gwr_regterm=gwr_regterm))
        flow.register_work(work)

    # RPA with quartic scaling. See also https://docs.abinit.org/tests/v67mbpt/Input/t19.abi
    work = flow.new_work()
    for gwr_ntau in gwr_ntau_list:
        chi_input = scf_input.new_with_vars(optdriver=3,       # Screening calculation.
                                            gwcalctyp=1,       # Gauss-Legendre frequency mesh on the imaginary axis.
                                            gwrpacorr=1,       # Exact integration over the coupling constant.
                                            nfreqim=gwr_ntau,  # No. of points along the imaginary axis for chi0.
                                            nband=nband,
                                            ecuteps=ecuteps,
                                           )
        work.register_scr_task(chi_input, deps={wfk_node: "WFK"})

    return flow


@flowtk.flow_main
def main(options):
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
"""
Script to generate/analyze/plot ONCVPSP pseudopotentials.
"""
from __future__ import annotations

import sys
import os
import argparse
import shutil
import abipy.tools.cli_parsers as cli

from pprint import pformat
from monty.termcolor import cprint
from abipy.flowtk.pseudos import Pseudo
from abipy.ppcodes.ppgen import OncvGenerator
from abipy.ppcodes.oncv_parser import OncvParser
from abipy.ppcodes.oncv_plotter import OncvPlotter, oncv_make_open_notebook, MultiOncvPlotter


def _find_oncv_output(path: str) -> str:
    """
    Fix possible error in the specification of filepath when we want a `.out` file.
    Return output path.
    """
    if path.endswith(".out"): return path
    root, _ = os.path.splitext(path)
    new_path = root + ".out"
    if not os.path.exists(new_path):
        raise ValueError("Cannot find neither %s nor %s" % (path, new_path))
    cprint("Maybe you meant %s" % new_path, "yellow")
    return new_path


def oncv_notebook(options):
    """
    Generate jupyter notebook to plot data. Requires oncvpsp output file.
    """
    out_path = _find_oncv_output(options.filepath)
    return oncv_make_open_notebook(out_path, foreground=options.foreground,
                                   classic_notebook=options.classic_notebook,
                                   no_browser=options.no_browser)


def oncv_gnuplot(options):
    """
    Plot data with gnuplot.
    """
    out_path = _find_oncv_output(options.filepath)

    # Parse output file.
    onc_parser = OncvParser(out_path).scan()
    if not onc_parser.run_completed:
        cprint("oncvpsp output is not completed. Exiting", "red")
        return 1

    onc_parser.gnuplot()
    return 0


def oncv_print(options) -> int:
    """
    Parse oncvps output and print results to terminal.
    """
    out_path = _find_oncv_output(options.filepath)
    p = OncvParser(out_path).scan()
    if not p.run_completed:
        raise RuntimeError("ocnvpsp output file is not completed")

    print(p.to_string(verbose=options.verbose))
    return 0


def oncv_plot(options) -> int:
    """
    Plot data with matplotlib. Requires oncvpsp output file.
    """
    cli.customize_mpl(options)

    out_path = _find_oncv_output(options.filepath)
    plotter = OncvPlotter.from_file(out_path)

    #plotter.plotly_atan_logders().show()
    #return 0

    plotter.expose(slide_mode=options.slide_mode, slide_timeout=options.slide_timeout,
                   use_web=options.expose_web, verbose=options.verbose)
    return 0


def oncv_plot_pseudo(options) -> int:
    """
    Plot data with matplotlib. Requires pseudopotential file (UPF2 or pawxml).
    """
    cli.customize_mpl(options)

    pseudo = Pseudo.from_file(options.filepath)
    print(pseudo)

    exposer = "panel" if options.expose_web else "mpl"
    from abipy.tools.plotting import Exposer
    with Exposer.as_exposer(exposer) as e:
        e.add_obj_with_yield_figs(pseudo)

    return 0


def oncv_compare(options) -> int:
    """
    Compare multiple oncvpsp output files.
    """
    cli.customize_mpl(options)

    out_paths = [_find_oncv_output(p) for p in options.filepaths]
    plotter = MultiOncvPlotter.from_files(out_paths)

    # Plot data
    plotter.expose(slide_mode=options.slide_mode, slide_timeout=options.slide_timeout,
                   use_web=options.expose_web, verbose=options.verbose)

    return 0


#@flowtk.flow_main
#def main(options):
#    return build_flow(options)

#def oncv_hints(options):
#    """
#    """
#    from abipy import flowtk
#    from abipy.flowtk.pseudo_works import GsEcutConvWork
#    flow = flowtk.Flow(workdir=options.workdir)
#
#    ecut_list = [35, 40, 45, 50, 55, 60, 65]
#    #ecut_list = [35, 40, 45]
#    work = GsEcutConvWork.from_pseudo(pseudo, ecut_list)
#    flow.register_work(work)
#    return flow


def oncv_run(options):
    """
    Run oncvpsp, generate djrepo file, plot results. Requires oncvps input file.
    """
    # Build names of psp8 and djson files from input and relativistic mode.
    in_path = options.filepath
    root, _ = os.path.splitext(in_path)

    # Enforce convention on output files.
    calc_type = None
    if options.rel == "nor":
        if not root.endswith("_nor"): root += "_nor"

    elif options.rel == "fr":
        if not root.endswith("_r"):
            root += "_r"
            cprint("FR calculation with input file without `_r` suffix. Will add `_r` to output files", "yellow")

    elif options.rel == "from_file":
        calc_type  = "scalar-relativistic"
        if root.endswith("_r"): calc_type = "fully-relativistic"
        if root.endswith("_nor"): calc_type = "non-relativistic"

    # Build names of output files.
    psp8_path = root + ".psp8"
    djrepo_path = root + ".djrepo"
    out_path = root + ".out"

    if os.path.exists(psp8_path):
        cprint("%s already exists and will be overwritten" % os.path.relpath(psp8_path), "yellow")
    if os.path.exists(djrepo_path):
        cprint("%s already exists and will be overwritten" % os.path.relpath(djrepo_path), "yellow")
    if os.path.exists(out_path):
        cprint("%s already exists and will be overwritten" % os.path.relpath(out_path), "yellow")

    # Select calc_type
    if calc_type is None:
        calc_type = dict(nor="non-relativistic",
                         sr="scalar-relativistic",
                         fr="fully-relativistic")[options.rel]

    # Build Generator and start generation.
    psgen = OncvGenerator.from_file(in_path, calc_type, options.use_mgga, workdir=None)

    print(psgen.input_str)
    print("Using executable:\n\t", psgen.executable)
    print(f"Output files produced in directory:\n\t{psgen.workdir}")

    if retcode := psgen.start_and_wait() != 0:
        cprint("oncvpsp returned %s. Exiting" % retcode, "red")
        return retcode

    if psgen.status != psgen.S_OK:
        cprint(f"psgen.status = {psgen.status} != psgen.S_OK", "red")

        if psgen.parser.warnings:
            print(2 * "\n")
            print("List of WARNINGS:")
            for w in psgen.parser.warnings:
                print(w)

        if psgen.parser.errors:
            print(2 * "\n")
            print("List of ERRORS:")
            for e in psgen.parser.errors:
                print(e)

    # Transfer final output file.
    shutil.copy(psgen.stdout_path, out_path)

    # Parse the output file.
    onc_parser = OncvParser(out_path).scan()
    if not onc_parser.run_completed:
        cprint("oncvpsp output is not completed. Exiting", "red")
        return 1

    # Extract psp8 files from the oncvpsp output and write it to file.
    with open(psp8_path, "wt") as fh:
        fh.write(onc_parser.get_psp8_str())

    # Write UPF2 file if available.
    upf_str = onc_parser.get_upf_str()
    if upf_str is not None:
        with open(psp8_path.replace(".psp8", ".upf"), "wt") as fh:
            fh.write(upf_str)
    else:
        cprint("UPF2 file has not been produced. Use `both` in input file!", "red")

    pseudo = Pseudo.from_file(psp8_path)
    if pseudo is None:
        cprint("Cannot parse psp8 file: %s" % psp8_path, "red")
        return 1

    # Initialize and write djson file.
    # FIXME: This part can be removed when we migrate to the new lightweight testing
    # infrastructure with small json files.
    from pseudo_dojo.core.dojoreport import DojoReport
    report = DojoReport.empty_from_pseudo(pseudo, onc_parser.hints, devel=False)
    report.json_write()

    cli.customize_mpl(options)

    # Build the plotter
    plotter = onc_parser.get_plotter()
    plotter.expose(slide_mode=options.slide_mode, slide_timeout=options.slide_timeout,
                   use_web=options.expose_web, verbose=options.verbose)
    return 0


def oncv_gui(options):
    """
    Start a panel web app to generate pseudopotentials.
    """
    import panel as pn
    from abipy.panels.oncvpsp_gui import OncvGui
    from abipy.panels.core import abipanel, get_abinit_template_cls_kwds, AbipyParameterized

    # Load abipy/panel extensions and set the default template
    abipanel(panel_template=options.panel_template)

    if options.has_remote_server:
        print("Enforcing limitations on what the user can do on the abinit server")
        AbipyParameterized.has_remote_server = options.has_remote_server

    tmpl_cls, tmpl_kwds = get_abinit_template_cls_kwds()

    if options.verbose:
        print("Using default template:", tmpl_cls, "with kwds:\n", pformat(tmpl_kwds), "\n")

    def build():
        gui = OncvGui.from_file(os.path.abspath(options.filepath), plotlyFlag=options.plotly)
        return tmpl_cls(main=gui.get_panel(), title="Oncvpsp GUI", **tmpl_kwds)

    # Call pn.serve to serve the multipage app.
    serve_kwargs = cli.get_pn_serve_kwargs(options)

    return pn.serve(build, **serve_kwargs)




def get_epilog() -> str:
        return """\
Usage example:

    oncv.py run H.in                   ==> Run oncvpsp input file (scalar relativistic mode).
    oncv.py run H_r.in                 ==> Run oncvpsp input file (relativistic mode).
    oncv.py plot H.out                 ==> Use matplotlib to plot oncvpsp results for pseudo H.psp8.
    oncv.py plot H.out -ew             ==> Show matplotlib figures in browser.
    oncv.py gui H.in                   ==> Start a panel web app to generate pseudopotential.
    oncv.py print H.out                ==> Parse oncvps output and print results to terminal.
    oncv.py compare H.out other_H.out  ==> Compare multiple output files (supports -ew option as well)
    oncv.py gnuplot H.out              ==> Use gnuplot to plot oncvpsp results for pseudo H.psp8.
    oncv.py notebook H.out             ==> Generate jupyter notebook to plot oncvpsp results.
"""


def get_parser(with_epilog=False):

    def get_copts_parser(multi=False):
        # Parent parser implementing common options.
        p = argparse.ArgumentParser(add_help=False)
        p.add_argument('-v', '--verbose', default=0, action='count', # -vv --> verbose=2
                       help='Verbose, can be supplied multiple times to increase verbosity')

        p.add_argument('--loglevel', default="ERROR", type=str,
                       help="set the loglevel. Possible values: CRITICAL, ERROR (default), WARNING, INFO, DEBUG")

        if multi:
            p.add_argument('filepaths', nargs="+", help="List of files to compare.")
        else:
            p.add_argument('filepath', default="", help="Path to the input/output file")

        return p

    copts_parser = get_copts_parser(multi=False)

    # Parent parser for commands supporting MplExposer.
    plot_parser = argparse.ArgumentParser(add_help=False)
    cli.add_expose_options_to_parser(plot_parser)

    # Build the main parser.
    parser = argparse.ArgumentParser(epilog=get_epilog(), formatter_class=argparse.RawDescriptionHelpFormatter)
    from abipy.core.release import __version__
    parser.add_argument('-V', '--version', action='version', version=__version__)

    # Create the parsers for the sub-commands
    subparsers = parser.add_subparsers(dest='command', help='sub-command help', description="Valid subcommands")

    # Subparser for run command.
    p_run = subparsers.add_parser('run', parents=[copts_parser, plot_parser], help=oncv_run.__doc__)
    p_run.add_argument("--rel", default="from_file", help=("Relativistic treatment: `nor` for non-relativistic, "
        "`sr` for scalar-relativistic, `fr` for fully-relativistic. Default: `from_file` i.e. detected from file"))
    p_run.add_argument("--use-mgga", action='store_true', default=False, help="Produce mega-gga pseudo with oncvpspm.x")

    # Subparser for print command.
    p_print = subparsers.add_parser('print', parents=[copts_parser], help=oncv_print.__doc__)

    # Subparser for plot command.
    p_plot = subparsers.add_parser('plot', parents=[copts_parser, plot_parser], help=oncv_plot.__doc__)

    # Subparser for plot command.
    p_plot_pseudo = subparsers.add_parser('plot_pseudo', parents=[copts_parser, plot_parser],
                                          help=oncv_plot_pseudo.__doc__)

    # Subparser for compare command.
    copts_parser_multi = get_copts_parser(multi=True)
    p_compare = subparsers.add_parser('compare', parents=[copts_parser_multi, plot_parser],
                                      help=oncv_compare.__doc__)

    # notebook options.
    p_nb = subparsers.add_parser('notebook', parents=[copts_parser], help=oncv_notebook.__doc__)
    p_nb.add_argument('-nb', '--notebook', action='store_true', default=False, help="Open file in jupyter notebook")
    p_nb.add_argument('--classic-notebook', "-cnb", action='store_true', default=False,
                          help="Use classic jupyter notebook instead of jupyterlab.")
    p_nb.add_argument('--no-browser', action='store_true', default=False,
                          help=("Start the jupyter server to serve the notebook "
                                "but don't open the notebook in the browser.\n"
                                "Use this option to connect remotely from localhost to the machine running the kernel"))
    p_nb.add_argument('--foreground', action='store_true', default=False,
                          help="Run jupyter notebook in the foreground.")

    parents = [copts_parser, cli.pn_serve_parser(), plot_parser]

    # Subparser for gui command.
    p_gui = subparsers.add_parser('gui', parents=parents, help=oncv_gui.__doc__)

    # Subparser for gnuplot command.
    p_gnuplot = subparsers.add_parser('gnuplot', parents=[copts_parser], help=oncv_gnuplot.__doc__)

    # Subparser for hints command.
    #p_hints = subparsers.add_parser('hints', parents=[copts_parser], help=oncv_hints.__doc__)
    #p_hints.add_argument("pseudo_paths", nargs="+", type=str, help="Pseudopotential path.")
    #p_hints.add_argument("--ecut", type=float, required=True, help="Cutoff energy in Ha.")
    #p_hints.add_argument("-rc", "--vloc-rcut-list", nargs="+", default=None, type=float,
    #                    help="List of cutoff radii for vloc in Bohr.")
    #cli.add_expose_options_to_parser(p_hints)

    return parser


def main():

    def show_examples_and_exit(err_msg=None, error_code=1):
        """Display the usage of the script."""
        sys.stderr.write(get_epilog())
        if err_msg: sys.stderr.write("Fatal Error\n" + err_msg + "\n")
        sys.exit(error_code)

    parser = get_parser(with_epilog=True)

    # Parse command line.
    try:
        options = parser.parse_args()
    except Exception as exc:
        show_examples_and_exit(error_code=1)

    cli.set_loglevel(options.loglevel)

    # Use seaborn settings.
    if getattr(options, "seaborn", None):
        import seaborn as sns
        sns.set(context=options.seaborn, style='darkgrid', palette='deep',
                font='sans-serif', font_scale=1, color_codes=False, rc=None)

    # Dispatch
    return globals()["oncv_" + options.command](options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
r"""
Phonon Band structures (Plotly version)
=======================================

This example shows how to plot the phonon band structure of AlAs with plotly.
See tutorial/lesson_rf2.html
"""
from abipy.abilab import abiopen
import abipy.data as abidata

# Open the PHBST file produced by anaddb and get the phonon bands.
with abiopen(abidata.ref_file("trf2_5.out_PHBST.nc")) as ncfile:
    phbands = ncfile.phbands

# Read the Phonon DOS from the netcd file produced by anaddb (prtdos 2)
with abiopen(abidata.ref_file("trf2_5.out_PHDOS.nc")) as ncfile:
    phdos = ncfile.phdos

#%%
# plot phonon bands and DOS.
# `plot_phonons_lo_to.py` shows how to treat the LO-TO splitting.

phbands.plotly(title="AlAs Phonon bands and DOS in eV")

#%%
# plot phonon bands with DOS.

phbands.plotly_with_phdos(phdos, units="cm-1", title="AlAs Phonon bands + DOS in cm-1")

#%%
# plot phonon DOS.

phdos.plotly(units="cm-1", title="Phonon DOS and IDOS in cm-1")

# Plot the phonon band structure with different color for each line.
#phbands.plot_colored_matched(units="cm-1",
#                             title="AlAs with different color for each line.")

# sphinx_gallery_thumbnail_path = '_static/plotly_logo.png'



""""Panels to interact with AbiPy flows."""
from __future__ import annotations

import textwrap
import traceback

import pandas as pd
import param
import bokeh.models.widgets as bkw
import panel as pn
import panel.widgets as pnw

from abipy.panels.core import AbipyParameterized, Loading, ButtonContext, depends_on_btn_click, dfc, ply
from abipy.flowtk.nodes import Node
#from abipy import flowtk


class FilePathSelect(pnw.Select):

    @classmethod
    def from_filepaths(cls, filepaths: list[str], 
                       filter_files=True, **kwargs):
        import os
        items = [(os.path.basename(p), p) for p in filepaths ]

        if filter_files:
            def filter_basename(name):
                if name.startswith(".") or name.endswith(".pickle"):
                    return False
                return True

            items = [t for t in items if filter_basename(t[0])]

        base2path = dict(items)
        new = cls(options=list(base2path.keys()), **kwargs)
        new._base2path = base2path

        return new

    @property
    def filepath(self) -> str:
        return self._base2path[self.value]

    def __bool__(self):
        return bool(self._base2path)

class NodeParameterized(AbipyParameterized):
    """

    """

    def __init__(self, node: Node, **params):
        super().__init__(**params)
        self.node = node

        if node.is_flow:
            self.flow = node
            self.nids = []
            for work in node:
                self.nids.append(work.node_id)
                self.nids.extend([task.node_id for task in work])

        elif node.is_work:
            self.flow = self.node.flow
            self.nids = [task.node_id for task in node]

        elif node.is_task:
            self.flow = self.node.flow
            self.nids = self.node.node_id

        else:
            raise ValueError(f"Don't know how to handle type: `{type(node)}`")

        self.engine = pnw.Select(value="fdp", name="engine",
                                 options=['dot', 'neato', 'twopi', 'circo', 'fdp', 'sfdp', 'patchwork', 'osage'])
        self.dirtree = pnw.Checkbox(name='Dirtree', value=False)
        self.graphviz_btn = pnw.Button(name="Show Graph", button_type='primary')

        self.status_btn = pnw.Button(name="Show Status", button_type='primary')

        self.history_btn = pnw.Button(name="Show history", button_type='primary')
        self.debug_btn = pnw.Button(name="Debug", button_type='primary')
        self.events_btn = pnw.Button(name="Show Events", button_type='primary')
        self.corrections_btn = pnw.Button(name="Show Corrections", button_type='primary')
        self.handlers_btn = pnw.Button(name="Show Handlers", button_type='primary')
        self.vars_text = pnw.TextInput(name='Abivars',
                                       placeholder='Enter list of variables separated by comma e.g. `ecut, natom`')
        self.vars_btn = pnw.Button(name="Show Variables", button_type='primary')
        #self.dims_btn = pnw.Button(name="Show Dimensions", button_type='primary')

        self.workdir_fileselector = pnw.FileSelector(node.workdir, only_files=True)
        self.outdir_fileselector = pnw.FileSelector(node.outdir.path)
        self.indir_fileselector = pnw.FileSelector(node.indir.path)

        # Create select widgets with the files in indir/outdir/workdir
        # Use basenames as items but remember that we need to abspath when opening the file.
        from abipy.flowtk.utils import Directory
        self.filepath_select_dir = {}
        for where in ("indir", "outdir", "workdir"):
            directory = Directory(self.node.workdir) if where == "workdir" else getattr(self.node, where)
            filepaths = directory.list_filepaths()
            self.filepath_select_dir[where] = FilePathSelect.from_filepaths(filepaths) #, name=f"Files in {where}")

    def get_status_view(self) -> pn.Column:
        return pn.Column(
            f"## Show the status of: `{repr(self.node)}`",
            pn.Row(
                self.wdg_box(["verbose", "status_btn"]),
                bkw.PreText(text=self.node.str_deps()),
            ),
            pn.layout.Divider(),
            self.on_status_btn,
            sizing_mode='stretch_width',
        )

    @depends_on_btn_click("status_btn")
    def on_status_btn(self) -> None:
        """
        Show the status of the node.
        """
        term = pnw.Terminal(
            output="\n\n",
            height=1200, # Need this one else the terminal is not shown properly
            sizing_mode='stretch_width',
        )
        term.write("\n")

        if self.node.is_task:
            return term

        # Here it's important to enforce verbose 1 else show_status
        # does not analyze the tasks that are completed.
        df = self.flow.show_status(nids=self.nids, stream=term, verbose=1, return_df=True) #self.verbose)

        return pn.Column(
                StatusCards(df),
                sizing_mode="stretch_width",
        )

        # TODO: Finalize the implementation.
        # Generate heatmap with plotly
        max_num_tasks = max(len(work) for work in self.flow)
        y = [f"w{i}" for i in range(len(self.flow))]
        x = [f"t{i}" for i in range(max_num_tasks)]
        z = []
        for work in self.flow:
            row = [None for i in range(max_num_tasks)]
            for i, task in enumerate(work):
                #row[i] = task.status
                row[i] = task.mpi_procs
            z.append(row)

        import plotly.graph_objects as go
        fig = go.Figure(data=go.Heatmap(
                        x=x, y=y, z=z,
                        hoverongaps=False, transpose=False, colorscale="Viridis",
                        ))
        fig.update_layout(title_text="Number of MPI procs", title_x=0.5)

        from abipy.tools.plotting import add_colorscale_dropwdowns
        add_colorscale_dropwdowns(fig)

        return ply(fig)

    def get_history_view(self) -> pn.Column:
        return pn.Column(
            f"## Show the history of: `{repr(self.node)}`",
            self.wdg_box(["verbose", "history_btn"]),
            pn.layout.Divider(),
            self.on_history_btn,
            sizing_mode='stretch_width',
       )

    @depends_on_btn_click("history_btn")
    def on_history_btn(self) -> None:
        """
        Show the history of the node.
        """
        term = pnw.Terminal(
            output="\n\n",
            height=1200, # Need this one else the terminal is not show properly
            sizing_mode='stretch_width',
        )

        self.flow.show_history(nids=self.nids,
                               stream=term,
                               #status=options.task_status,
                               #full_history=options.full_history,
                               #metadata=options.metadata
                               )
        return term

    def get_graphviz_view(self) -> pn.Column:
        return pn.Column(
                f"## Graphviz options for node: `{repr(self.node)}`",
                pn.WidgetBox(self.engine, self.dirtree, self.graphviz_btn),
                pn.layout.Divider(),
                self.on_graphviz_btn,
                sizing_mode="stretch_width"
        )

    @depends_on_btn_click("graphviz_btn")
    def on_graphviz_btn(self) -> None:
        """
        Visualize node dependencies with [graphviz package](https://graphviz.readthedocs.io/en/stable/index.html)
        """
        if self.dirtree.value:
            graph = self.node.get_graphviz_dirtree(engine=self.engine.value)
        else:
            graph = self.node.get_graphviz(engine=self.engine.value)

        #self.flow.plot_networkx(mode="network", with_edge_labels=False, ax=None, arrows=False,
        #                        node_size="num_cores", node_label="name_class", layout_type="spring", **kwargs):

        return pn.Column(
            "## Dependency Graph:",
            pn.pane.SVG(graph),
            sizing_mode="stretch_width"
        )

    def get_debug_view(self) -> pn.Column:
        return pn.Column(
            f"## Debug node:`{repr(self.node)}`",
            self.pws_col(["verbose", "debug_btn"]),
            self.on_debug_btn,
            pn.layout.Divider(),
            sizing_mode='stretch_width',
        )

        #d["Corrections"] = pn.Row(self.corrections_btn, self.on_corrections_btn)
        #d["Handlers"] = pn.Row(self.handlers_btn, self.on_handlers_btn)

    @depends_on_btn_click("debug_btn")
    def on_debug_btn(self) -> None:
        term = pnw.Terminal(output="\n\n",
            height=1200, # Need this one else the terminal is not show properly
            sizing_mode='stretch_width',
        )
        self.flow.debug(stream=term, nids=self.nids) # status=options.task_status,
        return term

    def get_events_view(self) -> pn.Column:
        return pn.Column(
            f"## Show the events of: `{repr(self.node)}`",
            self.pws_col(["verbose", "events_btn"]),
            self.on_events_btn,
            pn.layout.Divider(),
            sizing_mode='stretch_width',
        )

    @depends_on_btn_click("events_btn")
    def on_events_btn(self) -> pnw.Terminal:
        term = pnw.Terminal(
            output="\n\n",
            height=1200, # Need this one else the terminal is not show properly
            sizing_mode='stretch_width',
        )
        self.flow.show_events(stream=term, nids=self.nids) # status=options.task_status,
        return term

    @depends_on_btn_click("corrections_btn")
    def on_corrections_btn(self) -> pnw.Terminal:
        term = pnw.Terminal(
            output="\n\n",
            height=1200, # Need this one else the terminal is not show properly
            sizing_mode='stretch_width',
        )
        self.flow.show_corrections(stream=term, nids=self.nids)
        return term

    @depends_on_btn_click("handlers_btn")
    def on_handlers_btn(self):
        term = pnw.Terminal(
            output="\n\n",
            height=1200, # Need this one else the terminal is not show properly
            sizing_mode='stretch_width',
        )
        self.flow.show_event_handlers(stream=term, verbose=self.verbose) #, nids=self.nids,  status=None,
        return term

    def get_dims_and_vars_view(self) -> pn.Row:
        row = pn.Row(pn.Column(self.vars_text, self.vars_btn), self.on_vars_btn)
        return row
        #d["Dims"] = pn.Row(pn.Column(self.dims_btn), self.on_dims_btn)

    @depends_on_btn_click("vars_btn")
    def on_vars_btn(self):
        if not self.vars_text.value: return
        varnames = [s.strip() for s in self.vars_text.value.split(",")]
        df = self.flow.compare_abivars(varnames=varnames, nids=self.nids,
                                       printout=False, with_colors=False)
        return pn.Row(dfc(df))

    @depends_on_btn_click("dims_btn")
    def on_dims_btn(self) -> pn.Row:
        df = self.flow.get_dims_dataframe(nids=self.nids, printout=False, with_colors=False)
        return pn.Row(dfc(df), sizing_mode="scale_width")

    def get_files_in_dir_view(self, where):
        """
        Return None if no file is found
        """
        select = self.filepath_select_dir[where]
        if not select: return None

        btn = pnw.Button(name="Analyze", button_type='primary')
        output_area = pn.Column(sizing_mode="stretch_width")

        from abipy.abilab import abiopen
        #from .core import NcFileViewer

        def update_output_area(event):
            with ButtonContext(btn), Loading(output_area):
                try:
                    # Cannot close the file at this level because it may be needed by the new app.
                    abifile = abiopen(select.filepath)
                    output_area.objects = [abifile.get_panel()]
                except Exception as exc:
                    #print(exc)
                    #if select.filepath.endswith(".nc"):
                    #    # We have a nc file but it's not supported by abiopen.
                    #    # Let's create a minimalistic view of the netcdf dims/vars
                    #    #abifile = AbinitNcFile(select.filepath)
                    #    NcFileViewer(self).get_ncfile_view(**kwargs)
                    #    output_area.objects = [abifile.get_ncfile_view()]
                    #else:
                    obj = pn.pane.Markdown("```shell\n%s\n```" % traceback.format_exc())
                    output_area.objects = [obj]

        btn.on_click(update_output_area)

        return pn.Column(
            "## Select a file and click the button to analyze the data",
            pn.WidgetBox(select, btn),
            pn.layout.Divider(),
            output_area,
            sizing_mode="stretch_width"
        )

    #def on_workdir_selector_btn(self, event):
    #    """hello word"""
    #    filepaths = self.workdir_fileselector.value
    #    if not filepaths:
    #        objects = [pn.pane.Alert("## No file selected", altert_type="warning")]

    #    else:
    #        from abipy.abilab import abiopen
    #        objects = []
    #        for path in filepaths:
    #            try:
    #                abifile = abiopen(path)
    #                pn_obj = abifile.get_panel()
    #            except Exception:
    #                pn_obj = pn.pane.Alert(str(exc), altert_type="warning")

    #            objects.append(pn_obj)

    #    self.workdir_selector_output_area.objects = objects

    def get_panel(self, as_dict=False, **kwargs):
        """
        Return tabs with widgets to interact with the flow.
        """
        d = {}

        d["Status"] = self.get_status_view()
        d["History"] = self.get_history_view()
        d["Events"] = self.get_events_view()
        for where in ("workdir", "outdir", "indir"):
            if self.filepath_select_dir[where]:
                view = self.get_files_in_dir_view(where)
                if view is not None: d[where.capitalize()] = view

        d["Debug"] = self.get_debug_view()

        if not self.node.is_task:
            d["Dims & Vars"] = self.get_dims_and_vars_view()
        d["Graphviz"] = self.get_graphviz_view()

        if as_dict: return d

        return self.get_template_from_tabs(d, template=kwargs.get("template", None), closable=False)


class StatusCards(param.Parameterized):

    def __init__(self, df: pd.DataFrame, **params):
        self.df = df
        super().__init__(**params)

        # For each work_idx, compute the min/max index
        # we need this dict in add_vrect to shadow the region associated to the work.
        self.w_start_stop = {}
        for w_idx in df["work_idx"].unique():
            lst = sorted(df.index[(df['work_idx'] == w_idx)].tolist())
            self.w_start_stop[w_idx] = (lst[0], lst[-1])

        header_list = [
            "## DataFrame",
            "## Task status histogram",
            "## Task Timeline",
            "## Task class histogram",
            "## Runtime in seconds for each task in the flow (-1 if task is not running)",
            "## Number of WARNINGs found in the log file of the Task",
            "## Barplot with number of MPI procs",
        ]

        #if len(df['status'].unique()) != 1:
        # Show histogram with task status only if we have different status values.

        self.cards, self.done = {}, {}
        for header in header_list:
            card = pn.layout.Card(None, header=header, collapsed=True, sizing_mode="stretch_width")
            # Compute stuff only when user opens the card.
            card.param.watch(self.update_card, ['collapsed'])
            self.cards[header] = card
            self.done[header] = False

        # Compute this card
        self.cards["## Task status histogram"].collapsed = False

        open_btn = pnw.Button(name="Open all cards", button_type='primary')
        open_btn.on_click(self.open_all_cards)
        close_btn = pnw.Button(name="Close all cards", button_type='primary')
        close_btn.on_click(self.close_all_cards)

        self.layout = pn.Column(pn.Row(open_btn, close_btn),
                                *list(self.cards.values()),
                                sizing_mode="stretch_width"
                                )

    def __panel__(self):
        return self.layout

    def open_all_cards(self, event):
        for card in self.cards.values():
            card.collapsed = False

    def close_all_cards(self, event):
        for card in self.cards.values():
            card.collapsed = True

    def add_vrect_to_fig(self, fig):
        """
        Add vertical rectangles to the plotly fig in order to group tasks belonging to the same Work.
        """
        for w_idx, (x0, x1) in self.w_start_stop.items():
            fig.add_vrect(x0=x0, x1=x1,
                          annotation_text=f"w{w_idx}", annotation_position="top left",
                          fillcolor="grey", opacity=0.1, line_width=0)
        return fig

    def update_card(self, event):
        """
        This callback is triggered when the user opens/closes the card.
        Here we compute and display the plot/table the first time the card is opened.
        """
        header = event.obj.header
        if self.done[header]: return

        card = self.cards[header]
        has_pane = False
        import plotly.express as px
        df = self.df

        with Loading(card):
            if header == "## Task status histogram":
                fig = px.histogram(df, x="status")

            elif header == "## Task class histogram":
                fig = px.histogram(df, x="task_class", color="status")

            elif header == "## Runtime in seconds for each task in the flow (-1 if task is not running)":

                fig = px.scatter(df, x=df.index, y="task_runtime_s", color="status", #, symbol="work_idx", #size=
                                 hover_data =["num_warnings", "num_comments", "work_idx", "task_class"],
                                 hover_name="name")
                self.add_vrect_to_fig(fig)

            elif header == "## Number of WARNINGs found in the log file of the Task":

                fig = px.scatter(df, x=df.index, y="num_warnings", color="status", #, symbol="work_idx", #size=
                                 hover_data=["task_runtime_s", "num_comments", "work_idx", "task_class"],
                                 hover_name="name")
                self.add_vrect_to_fig(fig)

            elif header == "## Barplot with number of MPI procs":

                fig = px.bar(df, x="work_idx", y="mpi_procs", color="task_widx",
                             #barmode="group",
                             #color="status",
                             #pattern_shape="task_class",
                             #pattern_shape="status",
                             #pattern_shape_sequence=[".", "x", "+"],
                             hover_data =["num_warnings", "num_comments", "task_class"],
                             hover_name="name")

            elif header == "## Task Timeline":
                fig = px.timeline(df, x_start="start_datetime", x_end="end_datetime", y="name",
                                  color="task_class", hover_name="name")

            elif header == "## DataFrame":
                # Remove some columns as well as the index.
                simple_df = df.drop(columns=["node_id", "queue_id", "qname",
                                             "task_queue_time_s", "submission_datetime",
                                             "start_datetime", "end_datetime", "task_widx"]) # "work_idx",
                simple_df.set_index('name', inplace=True)

                pane = pnw.Tabulator(simple_df, groupby=['work_idx']) #, height=240)
                has_pane = True

            else:
                raise ValueError(f"No action registered for: `{header}`")

        if not has_pane:
            pane = ply(fig, with_divider=False)

        card.objects = [pane]
        self.done[header] = True



#!/usr/bin/env python
r"""
GWR flow with convergence studies
=================================

This script computes the G0W0 corrections with the GWR code
and the analytic continuation implemented in quartic scaling code in order to compare the results
"""

import os
import sys
import abipy.data as data
import abipy.abilab as abilab
import abipy.core.abinit_units as abu

from abipy import flowtk


def build_flow(options):

    from abipy.data.gwr_structures import get_gwr_structure
    symbol = "Si"
    #symbol = "LiF"
    #symbol = "SiC"
    #symbol = "C"
    #symbol = "BN"
    #symbol = "MgO"
    #symbol = "GaAs"
    #symbol = "ZnO"
    structure = get_gwr_structure(symbol)

    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_","flow_")

    from abipy.flowtk.psrepos import get_repo_from_name
    pseudos = get_repo_from_name("ONCVPSP-PBE-SR-PDv0.4").get_pseudos("standard")
    #pseudos = get_repo_from_name("ONCVPSP-PBE-SR-PDv0.4").get_pseudos("stringent")

    scf_input = abilab.AbinitInput(structure=structure, pseudos=pseudos)
    scf_input.set_cutoffs_for_accuracy("normal")
    ecut = scf_input["ecut"]
    scf_input.set_scf_nband_semicond()

    # Global variables.
    scf_input.set_vars(
        tolvrs=1e-8,
        paral_kgb=0,
        npfft=1,
        timopt=-1,
    )
    scf_input.set_kmesh(
        #ngkpt=[1, 1, 1],
        ngkpt=[2, 2, 2],
        #ngkpt=[4, 4, 4],
        #ngkpt=[8, 8, 8],
        shiftk=[0.0, 0.0, 0.0], # IMPORTANT: k-grid for GWR must be Gamma-centered.
    )

    # Get max number of PWs.
    #dims, _ = scf_input.abiget_dims_spginfo()
    #mpw = dims["mpw"]

    flow = flowtk.Flow(workdir=options.workdir)
    small_manager = options.manager.new_with_fixed_mpi_omp(4, 1)

    # GS-SCF run to get the DEN, followed by direct diago to obtain green_nband bands.
    from abipy.flowtk.gwr_works import DirectDiagoWork, GWRSigmaConvWork
    green_nband = -1  # -1 this means full diago
    diago_work = DirectDiagoWork.from_scf_input(scf_input, green_nband)
    diago_work[0].set_manager(small_manager)
    flow.register_work(diago_work)

    # Build template for GWR.
    ecuteps = 12
    nval = scf_input.num_valence_electrons // 2
    nband = nval * 100
    gwr_template = scf_input.make_gwr_qprange_input(gwr_ntau=6, nband=nband, ecuteps=ecuteps)

    # Define kptgw and bdgw
    kptgw = [ # k-points in reduced coordinates
        (0.0, 0.0, 0.0),
        (0.5, 0.5, 0.0), # X
        #(0.5    0.000    0.000),
    ]

    nkptgw = len(kptgw)
    bdgw = (nval, nval+1) * nkptgw

    sigma_kcalc_dict = dict(
        nkptgw=nkptgw,
        kptgw=kptgw,
        bdgw=bdgw,
    )

    gwr_template.set_vars(**sigma_kcalc_dict)

    gwr_ntau_list = list(range(6, 34, 2))
    gwr_ntau_list = [6, 8]
    gwr_ntau_list = [14, 20]

    # Conpute QP corrections without/with regularization term.
    # 1) Change the value of one variable:
    varname_values = ("gwr_ntau", gwr_ntau_list)
    gwr_template["nband"] = nband
    gwr_template["userra"] = 1e-6
    #gwr_template["userra"] = 0.0

    wfk_node = diago_work[1]
    gwr_work = GWRSigmaConvWork.from_varname_values(
            varname_values, gwr_template, den_node=diago_work[0], wfk_node=wfk_node)
    flow.register_work(gwr_work)

    # Create work for plasmon-pole calculation
    chi_ppm_input = scf_input.new_with_vars(optdriver=3,
                                            ecuteps=ecuteps,
                                            nband=nband,
                                            )

    sigma_ppm_input = chi_ppm_input.new_with_vars(optdriver=4,
                                                  ecutsigx=gwr_template["ecutsigx"],
                                                 )
    sigma_ppm_input.set_vars(**sigma_kcalc_dict)

    work = flow.new_work()
    scr_ppm_task = work.register_scr_task(chi_ppm_input, deps={wfk_node: "WFK"})
    work.register_sigma_task(sigma_ppm_input, deps={wfk_node: "WFK", scr_ppm_task: "SCR"})

    use_ac = False
    if use_ac:
        # Create work for G0W0 AC calculation
        #nfreqim  25
        #nomegasi  10
        #omegasimax 10 eV
        ac_work = flow.new_work()
        for gwr_ntau in gwr_ntau_list:
            chi_ac_input = scf_input.new_with_vars(optdriver=3,
                                                   gwcalctyp=1,       # Analytic continuation.
                                                   nfreqim=gwr_ntau,  # This is equal to gwr_ntau
                                                   ecuteps=ecuteps,
                                                   nband=nband,
                                                   )

            sigma_ac_input = chi_ac_input.new_with_vars(optdriver=4,
                                                        nomegasi=gwr_ntau,
                                                        omegasimax=0.2 * abu.eV_Ha * gwr_ntau,
                                                        ecutsigx=gwr_template["ecutsigx"],
                                                       )
            sigma_ac_input.set_vars(**sigma_kcalc_dict)
            scr_ac_task = ac_work.register_scr_task(chi_ac_input, deps={wfk_node: "WFK"})
            ac_work.register_sigma_task(sigma_ac_input, deps={wfk_node: "WFK", scr_ac_task: "SCR"})

    use_cd = True
    if use_cd:
        # Create work for G0W0 with contour deformation.
        cd_work = flow.new_work()
        chi_cd_input = scf_input.new_with_vars(optdriver=3,
                                               gwcalctyp=2,       # Contour deformation
                                               nfreqre=50,
                                               freqremax=1.5,
                                               nfreqim=10,
                                               ecuteps=ecuteps,
                                               nband=nband,
                                               )

        sigma_cd_input = chi_cd_input.new_with_vars(optdriver=4,
                                                    ecutsigx=gwr_template["ecutsigx"],
                                                    #nfreqsp=200,
                                                    #freqspmax=2.,
                                                    )
        sigma_cd_input.set_vars(**sigma_kcalc_dict)
        scr_cd_task = cd_work.register_scr_task(chi_cd_input, deps={wfk_node: "WFK"})
        cd_work.register_sigma_task(sigma_cd_input, deps={wfk_node: "WFK", scr_cd_task: "SCR"})

    return flow


@flowtk.flow_main
def main(options):
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python
"""Electron-phonon calculations."""
from __future__ import print_function, division, unicode_literals, absolute_import

import os
import sys
import numpy as np
import abipy.data as abidata
import abipy.abilab as abilab
import abipy.flowtk as flowtk


def build_flow(options):
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    workdir = options.workdir
    if not options.workdir:
        workdir = os.path.basename(__file__).replace(".py", "").replace("run_", "flow_")

    # Preparatory run for E-PH calculations.
    # The sequence of datasets makes the ground states and
    # all of the independent perturbations of the single Al atom
    # for the irreducible qpoints in a 4x4x4 grid.
    # Note that the q-point grid must be a sub-grid of the k-point grid (here 8x8x8)
    pseudos = abidata.pseudos("Al.oncvpsp")

    structure = abilab.Structure.from_abivars(
        acell=3*[7.5],
        rprim=[0.0, 0.5, 0.5,
               0.5, 0.0, 0.5,
               0.5, 0.5, 0.0],
        typat=1,
        xred=[0.0, 0.0, 0.0],
        ntypat=1,
        znucl=13,
    )

    gs_inp = abilab.AbinitInput(structure, pseudos)

    gs_inp.set_vars(
        istwfk="*1",
        ecut=12.0,
        nband=5,
        occopt=7,    # include metallic occupation function with a small smearing
        tsmear=0.04,
        tolvrs=1e-7,
        timopt=-1,
    )

    # The kpoint grid is minimalistic to keep the calculation manageable.
    gs_inp.set_kmesh(
        ngkpt=[12, 12, 12],
        shiftk=[0.0, 0.0, 0.0],
        #kptopt=3,
    )

    flow = flowtk.Flow(workdir, manager=options.manager, remove=options.remove)

    # GS run to get the WFK
    work0 = flow.register_task(gs_inp, task_class=flowtk.ScfTask)

    # Phonon work with 4x4x4 q-mesh
    ddb_ngqpt = [4, 4, 4]
    ph_work = flowtk.PhononWork.from_scf_task(work0[0], ddb_ngqpt, is_ngqpt=True)
    flow.register_work(ph_work)

    # Build input file for E-PH run.
    eph_inp = gs_inp.new_with_vars(
        optdriver=7,
        ddb_ngqpt=ddb_ngqpt,       # q-mesh used to produce the DDB file (must be consistent with DDB data)
        eph_intmeth=2,             # Tetra method
        eph_fsewin="0.8 eV",       # Energy window around Ef
        eph_mustar=0.12,           # mustar parameter
    )

    # Set q-path for phonons and phonon linewidths.
    eph_inp.set_qpath(10)

    # Set q-mesh for phonons DOS and a2F(w)
    eph_inp.set_phdos_qmesh(nqsmall=16, method="tetra")

    # EPH part requires the GS WFK, the DDB file with all perturbations
    # and the database of DFPT potentials (already merged by PhononWork)
    eph_work = flow.register_work(flowtk.Work())
    deps = {work0[0]: "WFK", ph_work: ["DDB", "DVDB"]}

    eph_work.register_eph_task(eph_inp, deps=deps)

    # Activate Fourier interpolation of DFPT potentials.
    eph_work.register_eph_task(eph_inp.new_with_vars(eph_ngqpt_fine=[6, 6, 6]), deps=deps)
    eph_work.register_eph_task(eph_inp.new_with_vars(eph_ngqpt_fine=[12, 12, 12]), deps=deps)

    flow.allocate()

    # EPH does not support autoparal (yet)
    #for eph_task in eph_work:
    #    eph_task.with_fixed_mpi_omp(1, 1)

    return flow


@flowtk.flow_main
def main(options):
    return build_flow(options)


if __name__ == "__main__":
    retcode = main()
    if retcode != 0: sys.exit(retcode)

    rename_table = [
        #  src, dest
        ("_runflow/w2/t0/outdata/out_A2F.nc", "al_888k_161616q_A2F.nc"),
        ("_runflow/w2/t1/outdata/out_A2F.nc", "al_888k_161616q_vinterp_A2F.nc"),
	("_runflow/w2/t2/outdata/out_PHDOS.nc", "al_161616q_PHDOS.nc"),
    ]
    import shutil
    for old, new in rename_table:
        shutil.copyfile(old, new)
    #shutil.rmtree("_runflow")
    sys.exit(0)



#!/usr/bin/env python
r"""
e-ph matrix elements along a q-path
===================================

This example shows how to compute the e-ph matrix elements in AlAs along a q-path with AbiPy flows.
The final results are stored in the GKQ.nc file (one file for q-point) in the outdata of each task.
"""
import sys
import os
import abipy.abilab as abilab
import abipy.data as abidata

from abipy import flowtk


def make_scf_input(ngkpt):
    """
    This function constructs the input file for the GS calculation:
    """
    structure = abidata.structure_from_ucell("AlAs")
    pseudos = abidata.pseudos("13al.981214.fhi", "33as.pspnc")
    gs_inp = abilab.AbinitInput(structure, pseudos=pseudos)

    gs_inp.set_vars(
        nband=6,
        ecut=6.0,
        ngkpt=ngkpt,
        nshiftk=1,
        shiftk=[0, 0, 0],
        tolvrs=1.0e-10,
    )

    return gs_inp


def build_flow(options):
    """
    Create a `Flow` for the computation of e-ph matrix elements
    """
    # Working directory (default is the name of the script with '.py' removed and "run_" replaced by "flow_")
    if not options.workdir:
        options.workdir = os.path.basename(sys.argv[0]).replace(".py", "").replace("run_", "flow_")

    # Use 2x2x2 both for k-mesh.
    # Build input for GS calculation
    scf_input = make_scf_input(ngkpt=(2, 2, 2))

    # corresponding to a [2, 2, 2] q-mesh.
    ngqpt = (2, 2, 2)

    # Create flow to compute all the independent atomic perturbations
    # Use ndivsm = 0 to pass an explicit list of q-points.
    # If ndivsm > 0, qpath_list is interpreted as a list of boundaries for the q-path
    qpath_list = [[0.0, 0.0, 0.0], [0.01, 0, 0], [0.1, 0, 0],
                  [0.24, 0, 0], [0.3, 0, 0], [0.45, 0, 0], [0.5, 0.0, 0.0]]

    from abipy.flowtk.eph_flows import GkqPathFlow
    flow = GkqPathFlow.from_scf_input(options.workdir, scf_input,
                                      ngqpt, qpath_list, ndivsm=0, with_becs=True,
                                      ddk_tolerance={"tolwfr": 1e-8})

    return flow


# This block generates the thumbnails in the AbiPy gallery.
# You can safely REMOVE this part if you are using this script for production runs.
if os.getenv("READTHEDOCS", False):
    __name__ = None
    import tempfile
    options = flowtk.build_flow_main_parser().parse_args(["-w", tempfile.mkdtemp()])
    build_flow(options).graphviz_imshow()


@flowtk.flow_main
def main(options):
    """
    This is our main function that will be invoked by the script.
    flow_main is a decorator implementing the command line interface.
    Command line args are stored in `options`.
    """
    return build_flow(options)


if __name__ == "__main__":
    sys.exit(main())


############################################################################
#
# Run the script with:
#
#     run_gkq.py -s
#
# then use:
#
#    abirun.py flow_phonons history



#!/usr/bin/env python
r"""
Band structure plot
===================

This example shows how to plot a band structure
using the eigenvalues stored in the GSR file
produced at the end of the GS run.
"""
from abipy.abilab import abiopen
import abipy.data as abidata

#%%
# Here we use one of the GSR files shipped with abipy.
# Replace filename with the path to your GSR file or your WFK file.

filename = abidata.ref_file("si_nscf_GSR.nc")

#%%
# Open the GSR file and extract the band structure.
# (alternatively one can use the shell and `abiopen.py OUT_GSR.nc -nb`
# to open the file in a jupyter notebook.

with abiopen(filename) as ncfile:
    ebands = ncfile.ebands

#%%
# Plot the band energies with matplotlib.
# Note that the labels for the k-points are found automatically in an internal database.
# Use `with_gaps` to show fundamental and direct gaps.

ebands.plot(with_gaps=True, title="Silicon band structure")

#%%
# For the plotly version, use:

ebands.plotly(with_gaps=True, title="Silicon band structure")

# .. warning:
#
# Note that, for the time being, ``with_gaps`` is incompatible with the ``title`` argument.

#%%
# Plot the BZ and the k-point path with matplotlib

ebands.kpoints.plot()



# coding: utf-8
"""Utilities for working with strings and text."""
from __future__ import annotations


def tonumber(s: str) -> float:
    """
    Convert string to number, raise ValueError if s cannot be converted.
    """
    # Duck test.
    try:
        stnum = s.upper().replace("D", "E")  # D-01 is not recognized by python: Replace it with E.
        # stnum = strip_punct(stnum)         # Remove punctuation chars.
        return float(stnum)                  # Try to convert.

    except ValueError:
        raise

    except Exception:
        raise RuntimeError("Don't know how to handle type %s: %s" % (type(s), str(s)))


def nums_and_text(line: str) -> tuple:
    """
    Split line into (numbers, text).
    """
    tokens = line.split()
    text = ""
    numbers = []

    for tok in tokens:
        try:
            numbers.append(tonumber(tok))
        except ValueError:
            text += " " + tok

    return numbers, text


def rreplace(s, old, new, occurrence):
    """
    replace old with new in string but, instead of starting from the beginning
    as replace does, starting from the end.

    >>> s = '1232425'
    >>> assert rreplace(s, '2', ' ', 2) == '123 4 5'
    >>> assert rreplace(s, '2', ' ', 3) == '1 3 4 5'
    >>> assert rreplace(s, '2', ' ', 4) == '1 3 4 5'
    >>> assert rreplace(s, '2', ' ', 0) == '1232425'
    """
    # Based on:
    # https://stackoverflow.com/questions/2556108/rreplace-how-to-replace-the-last-occurrence-of-an-expression-in-a-string
    li = s.rsplit(old, occurrence)
    return new.join(li)


def rm_multiple_spaces(string: str) -> str:
    """remove multiple spaces in a string."""
    import re
    return re.sub(' +', ' ', string)



# coding: utf-8
"""
Interface to the PSPS.nc file containing the splined form factors computed by ABINIT.
"""
from __future__ import annotations

import os
import numpy as np
import pandas as pd

#from typing import List, Any
from collections import OrderedDict
from monty.bisect import find_gt
from monty.string import marquee # list_strings,
from monty.functools import lazy_property
from abipy.iotools import ETSF_Reader
from abipy.core.structure import Structure
from abipy.core.mixins import AbinitNcFile, NotebookWriter
from abipy.abio.robots import Robot
from abipy.tools.plotting import add_fig_kwargs, get_ax_fig_plt, get_axarray_fig_plt, set_visible
from abipy.tools.typing import Figure


def _mklabel(fsym: str, der: int, arg: str) -> str:
    """
    mklabel(f, 2, x) --> $f''(x)$
    """
    if der == 0:
        return "$%s(%s)$" % (fsym, arg)
    else:
        fsym = fsym + "^{" + (der * r"\prime") + "}"
        return "$%s(%s)$" % (fsym, arg)


def _rescale(arr, scale=1.0):
    if scale is None:
        return arr, 0.0

    amax = np.abs(arr).max()
    fact = scale / amax if amax != 0 else 1
    return fact * arr, fact


def dataframe_from_pseudos(pseudos, index=None) -> pd.DataFrame:
    """
    Build pandas dataframe with the most important info associated to
    a list of pseudos or a list of objects that can be converted into pseudos.

    Args:
        pseudos: List of objects that can be converted to pseudos.
        index: Index of the dataframe.

    Return: pandas Dataframe.
    """
    from abipy.flowtk import PseudoTable
    pseudos = PseudoTable.as_table(pseudos)

    attr_names = ["Z_val", "l_max", "l_local", "nlcc_radius", "xc", "supports_soc", "type"]
    rows = []
    for p in pseudos:
        row = OrderedDict([(k, getattr(p, k, None)) for k in attr_names])
        row["ecut_normal"], row["pawecutdg_normal"] = None, None
        if p.has_hints:
            hint = p.hint_for_accuracy(accuracy="normal")
            row["ecut_normal"] = hint.ecut
            if hint.pawecutdg: row["pawecutdg_normal"] = hint.pawecutdg
        rows.append(row)

    return pd.DataFrame(rows, index=index, columns=list(rows[0].keys()) if rows else None)


_LW = 1.0

_Q2VQ_LATEX = r"$f(q) = q^2 V(q) = -\frac{Z_v}{\pi} + q^2 4\pi\int (\frac{\sin(2\pi q r)}{2\pi q r})(r^2 V(r)+r Z_v) dr$"


class PspsFile(AbinitNcFile, NotebookWriter):
    """
    Netcdf file with the tables used in Abinit to apply the pseudopotential part of the KS Hamiltonian.

    Usage example:

    .. code-block:: python

        with PspsFile("foo_PSPS.nc") as psps:
            psps.plot_tcore_rspace()
    """
    linestyles_der = ["-", "--", '-.', ':', ":", ":"]

    color_der = ["black", "red", "green", "orange", "cyan"]

    @classmethod
    def from_file(cls, filepath: str) -> PspsFile:
        """Initialize the object from a filepath."""
        return cls(filepath)

    @classmethod
    def from_abinit_run(cls, pseudo, ecut, vloc_rcut=None, workdir=None) -> PspsFile:
        """
        Initialize the object from a filepath or a Pseudo object by invoking
        abinit with cutoff energy `ecut`.

        Args:
            vloc_rcut: Radial cutoff in Bohr (Abinit input variable)
        """
        from abipy.flowtk.pseudos import Pseudo
        pseudo = Pseudo.as_pseudo(pseudo)
        structure = Structure.boxed_atom(pseudo)

        from abipy.abio.factories import gs_input
        inp = gs_input(structure, pseudo,
                       kppa=1, ecut=ecut, pawecutdg=None, accuracy="normal", spin_mode="unpolarized",
                       smearing="fermi_dirac:0.1 eV", charge=0.0)

        inp.set_vars(prtpsps=-1,       # Print PSPS.nc and exit immediately
                     vloc_rcut=vloc_rcut,
                     )

        from abipy.flowtk import AbinitTask
        task = AbinitTask.temp_shell_task(inp, workdir=workdir)
        retcode = task.start_and_wait(autoparal=False) #, exec_args=["--dry-run"])
        #if retcode != 0:
        #    print("task.workdir:", task.workdir, "returned retcode", retcode)

        return cls(os.path.join(task.outdir.path_in("out_PSPS.nc")))

    def __init__(self, filepath: str):
        super().__init__(filepath)
        self.reader = PspsReader(filepath)

        # TODO
        #self.ecut = self.reader.read_value("ecut")

    def close(self) -> None:
        """Close the file."""
        self.reader.close()

    @lazy_property
    def params(self) -> dict:
        """:class:`OrderedDict` with parameters that might be subject to convergence studies."""
        return {}

    def __str__(self) -> str:
        """String representation."""
        return self.to_string()

    def to_string(self, verbose: int = 0) -> str:
        """String representation."""
        lines = []; app = lines.append
        app(marquee("File Info", mark="="))
        app(self.filestat(as_string=True))
        app("")

        if verbose > 1:
            app("")
            #app(self.hdr.to_string(verbose=verbose, title="Abinit Header"))

        return "\n".join(lines)

    @add_fig_kwargs
    def plot(self, **kwargs) -> Figure:
        """
        Driver routine to plot several quantities on the same graph.

        Args:
            ecut_ffnl: Max cutoff energy for ffnl plot (optional)

        Return: |matplotlib-Figure|
        """
        methods = [
            "plot_tcore_rspace",
            "plot_tcore_qspace",
            "plot_ffspl",
            "plot_q2vq",
        ]

        ax_list, fig, plt = get_axarray_fig_plt(None, nrows=2, ncols=2,
                                                sharex=False, sharey=False, squeeze=True)

        ecut_ffnl = kwargs.pop("ecut_ffnl", None)
        for m, ax in zip(methods, ax_list.ravel()):
            getattr(self, m)(ax=ax, ecut_ffnl=ecut_ffnl, show=False)

        return fig

    @add_fig_kwargs
    def plot_tcore_rspace(self, ax=None, ders=(0, 1, 2, 3), scale=1.0, rmax=3.0, **kwargs) -> Figure:
        """
        Plot the model core charge and its derivatives in real space.

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.
            ders: Tuple selecting the derivatives to be plotted.
            scale: 1.0 if all derivatives should be scaled to 1 else None.
            rmax: Max radius for plot in Bohr. None is full grid is wanted.

        Returns: |matplotlib-Figure|
        """
        if not isinstance(ders, (list, tuple)): ders = [ders]

        rmeshes, coresd = self.reader.read_coresd(rmax=rmax)
        ax, fig, plt = get_ax_fig_plt(ax=ax)

        for rmesh, mcores in zip(rmeshes, coresd):
            for der, values in enumerate(mcores):
                if der not in ders: continue
                yvals, fact, = _rescale(values, scale=scale)
                ax.plot(rmesh, yvals,
                        color=kwargs.get("color", self.color_der[der]),
                        linewidth=kwargs.get("linewidth", _LW),
                        linestyle=kwargs.get("linestyle", self.linestyles_der[der]),
                        label=_mklabel(r"\tilde{n}_c", der, "r") + " x %.4f" % fact
                       )

        ax.grid(True)
        ax.set_xlabel("r (Bohr)")
        ax.set_title("Model core in r-space")
        if kwargs.get("with_legend", False):
            ax.legend(loc="best")

        return fig

    @add_fig_kwargs
    def plot_tcore_qspace(self, ax=None, ders=(0,), with_fact=True,
                          with_qn=0, scale=1.0, **kwargs) -> Figure:
        """
        Plot the model core charge in q space.

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.
            ders: Tuple selecting the derivatives to be plotted.
            scale: 1.0 if all derivatives should be scaled to 1 else None.
            with_qn:

        Returns: |matplotlib-Figure|
        """
        if not isinstance(ders, (list, tuple)): ders = [ders]

        ax, fig, plt = get_ax_fig_plt(ax=ax)

        color = kwargs.pop("color", "black")
        linewidth = kwargs.pop("linewidth", _LW)

        qmesh, tcore_spl = self.reader.read_tcorespl()
        ecuts = 2 * (np.pi * qmesh)**2

        lines = []
        for atype, tcore_atype in enumerate(tcore_spl):
            for der, values in enumerate(tcore_atype):
                #_ider = {0: 0, 1: 2}[der]
                if der not in ders: continue
                yvals, fact = _rescale(values, scale=scale)

                label = _mklabel("\\tilde{n}_{c}", der, "q")
                if with_fact: label += " x %.4f" % fact

                line, = ax.plot(ecuts, yvals, color=color, linewidth=linewidth,
                                linestyle=self.linestyles_der[der], label=label)
                lines.append(line)

                if with_qn and der == 0:
                    yvals, fact = _rescale(qmesh * values, scale=scale)
                    line, ax.plot(ecuts, yvals, color=color, linewidth=linewidth,
                                  label=_mklabel("q f", der, "q") + " x %.4f" % fact)
                    lines.append(line)

        ax.grid(True)
        ax.set_xlabel("Ecut (Hartree)")
        ax.set_title("Model core in q-space")
        if kwargs.get("with_legend", False):
            ax.legend(loc="best")

        return fig

    @add_fig_kwargs
    def plot_q2vq(self, ax=None, ders=(0,), with_qn=0, with_fact=True,
                  scale=None, **kwargs) -> Figure:
        r"""
        Plot the local part of the pseudopotential in q space.

        q2vq(mqgrid) = q^2 V(q)
         = -\frac{Zv}{\pi}
           + q^2 4\pi\int[(\frac{\sin(2\pi q r)}{2\pi q r})(r^2 V(r)+r Zv)dr].

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.
            ders: Tuple used to select the derivatives to be plotted.
            scale: 1.0 if all derivatives should be scaled to 1 else None.
            with_qn:

        Returns: |matplotlib-Figure|
        """
        if not isinstance(ders, (list, tuple)): ders = [ders]

        ax, fig, plt = get_ax_fig_plt(ax=ax)

        color = kwargs.pop("color", "black")
        linewidth = kwargs.pop("linewidth", _LW)

        qmesh, vlspl = self.reader.read_vlspl()
        ecuts = 2 * (np.pi * qmesh)**2
        for atype, vl_atype in enumerate(vlspl):
            for der, values in enumerate(vl_atype):
                if der not in ders: continue
                #_ider = {0: 0, 1: 2}[der]
                yvals, fact = _rescale(values, scale=scale)
                label = _mklabel("q^2 v_{loc}", der, "q")
                if with_fact: label += " x %.4f" % fact

                ax.plot(ecuts, yvals, color=color, linewidth=linewidth,
                        linestyle=self.linestyles_der[der], label=label)

                #if with_qn and der == 0:
                #    yvals, fact = _rescale(qmesh * values, scale=scale)
                #    ax.plot(ecuts, yvals, color=color, linewidth=linewidth, label="q*f(q) x %2.f" % fact)

                #if der == 0:
                #   z_val = self.reader.zion_typat[atype]
                #   ax.axhline(y=-z_val / np.pi, linewidth=1, color='k', linestyle="dashed")

        ax.grid(True)
        ax.set_xlabel("Ecut (Hartree)")
        ax.set_title("Vloc(q)")
        if kwargs.get("with_legend", False):
            ax.legend(loc="best")

        return fig

    @add_fig_kwargs
    def plot_ffspl(self, ax=None, ecut_ffnl=None, ders=(0,), l_select=None,
                   with_qn=0, with_fact=False, scale=None, **kwargs) -> Figure:
        """
        Plot the nonlocal part of the pseudopotential in q-space.

        Args:
            ax: |matplotlib-Axes| or None if a new figure should be created.
            ecut_ffnl: Max cutoff energy for ffnl plot (optional)
            scale: 1.0 if all derivatives should be scaled to 1 else None.
            ders: Tuple used to select the derivatives to be plotted.
            with_qn:

        Returns: |matplotlib-Figure|
        """
        if not isinstance(ders, (list, tuple)): ders = [ders]
        if l_select is not None:
            if not isinstance(l_select, (list, tuple)): l_select = [l_select]

        ax, fig, plt = get_ax_fig_plt(ax=ax)

        color = kwargs.pop("color", "black")
        linewidth = kwargs.pop("linewidth", _LW)

        color_l = {-1: "black", 0: "red", 1: "blue", 2: "green", 3: "orange"}
        linestyles_n = ["solid", '-', '--', '-.', ":"]
        l_seen = set()

        # vlspl has shape [ntypat, 2, mqgrid_vl]
        qmesh, vlspl = self.reader.read_vlspl()
        all_projs = self.reader.read_projectors()

        for itypat, projs_type in enumerate(all_projs):
            # Loop over the projectors for this atom type.
            for p in projs_type:
                if l_select is not None and p.l not in l_select: continue
                print("Printing:", p)

                for der, values in enumerate(p.data):
                    if der not in ders: continue
                    #_ider = {0: 0, 1: 2}[der]
                    #yvals, fact = _rescale(values, scale=scale)

                    label = None
                    if p.l not in l_seen:
                        l_seen.add(p.l)
                        label = _mklabel("v_{nl}", der, "q") + ", l=%d" % p.l

                    stop = len(p.ecuts) +  1
                    if ecut_ffnl is not None:
                        stop = find_gt(p.ecuts, ecut_ffnl)

                    #values = p.ekb * p.values - vlspl[itypat, 0, :]
                    #values = vlspl[itypat, der] + p.sign_sqrtekb * p.values
                    #values = p.sign_sqrtekb * p.values
                    values = p.data[der]

                    #print(values.min(), values.max())
                    ax.plot(p.ecuts[:stop], values[:stop],
                            color=color_l[p.l], linewidth=linewidth,
                            linestyle=linestyles_n[p.n], label=label)

        ax.grid(True)
        ax.set_xlabel("Ecut (Hartree)")
        #ax.set_title("ffnl(q)")
        if kwargs.get("with_legend", False):
            ax.legend(loc="best")

        #ax.axvline(x=self.ecut, linewidth=linewidth, color='k', linestyle="solid")
        #ax.axhline(y=0, linewidth=linewidth, color='k', linestyle="solid")

        fig.tight_layout()

        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        yield self.plot(show=False)

    def write_notebook(self, nbpath=None):
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        nb.cells.extend([
            #nbv.new_markdown_cell("# This is a markdown cell"),
            nbv.new_code_cell("psps = abilab.abiopen(%s)" % self.filepath),
        ])

        return self._write_nb_nbpath(nb, nbpath)


class PspsRobot(Robot):
    """
    This robot analyzes the results contained in multiple PSPS.nc files.

    .. rubric:: Inheritance Diagram
    .. inheritance-diagram:: PspsRobot
    """

    EXT = "PSPS"

    @classmethod
    def from_abinit_run(cls, pseudos, ecut) -> PspsRobot:
        """
        Initialize the object from a list of filepaths or Pseudo objects
        """
        filepaths = []
        for p in pseudos:
            with PspsFile.from_abinit_run(p, ecut) as psps:
                filepaths.append(psps.filepath)

        return cls.from_files(filepaths)

    @classmethod
    def from_vloc_rcut_list(cls, pseudo, vloc_rcut_list, ecut) -> PspsRobot:
        """
        Initialize the object by computing the form factors for the same pseudo
        with different values of `vloc_rcut_list` in Bohr.
        """
        filepaths = []
        for vloc_rcut in vloc_rcut_list:
            with PspsFile.from_abinit_run(pseudo, ecut, vloc_rcut=vloc_rcut) as psps:
                filepaths.append(psps.filepath)

        return cls.from_files(filepaths)

    def _mkcolor(self, count, cmap="jet"):
        npseudos = len(self)
        if npseudos <= 2:
            return {0: "red", 1: "blue", 2: "green"}[count]
        else:
            import matplotlib.pyplot as plt
            cmap = plt.get_cmap(cmap)
            return cmap(float(count) / (npseudos - 1))

    @add_fig_kwargs
    def plot_tcore_rspace(self, ders=(0, 1, 2, 3), scale=None, fontsize=8, **kwargs) -> Figure:
        """
        Plot the model core charge and its derivatives in r-space.

        Args:
            fontsize: fontsize for subtitles.

        Returns: |matplotlib-Figure|
        """
        nrows, ncols = len(self), len(ders)
        ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                               sharex=True, sharey=False, squeeze=False)

        fig.suptitle(f"Model core in r-space")
        for i, (label, psps) in enumerate(self.items()):
            kws = dict(color=self._mkcolor(i), show=False)
            for j, der in enumerate(ders):
                ax = ax_mat[i,j]
                psps.plot_tcore_rspace(ax=ax, ders=der, with_legend=False, scale=scale, **kws)
                ax.set_title(f"$rho_M^{der}(r)$", fontsize=fontsize) if i == 0 else ax.set_title("")
                if i != len(self) - 1: set_visible(ax, False, "xlabel")

        return fig

    @add_fig_kwargs
    def plot_tcore_qspace(self, ders=(0, 1), with_qn=0, scale=None,
                          fontsize=8, **kwargs) -> Figure:
        """
        Plot the model core charge and its derivatives in q-space.

        Args:
            fontsize: fontsize for subtitles.

        Returns: |matplotlib-Figure|
        """
        nrows, ncols = len(self), len(ders)
        ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                               sharex=True, sharey=False, squeeze=False)

        fig.suptitle(f"Model core in q-space")
        for i, (label, psps) in enumerate(self.items()):
            kws = dict(color=self._mkcolor(i), show=False)
            for j, der in enumerate(ders):
                #_ider = {0: 0, 1: 2}[der]
                ax = ax_mat[i,j]
                psps.plot_tcore_qspace(ax=ax, ders=der, with_qn=with_qn, scale=scale, **kws)
                ax.set_title(f"$rho_M^{der}(q)$", fontsize=fontsize) if i == 0 else ax.set_title("")
                if i != len(self) - 1: set_visible(ax, False, "xlabel")

        return fig

    @add_fig_kwargs
    def plot_q2vq(self, ders=(0, 1), with_qn=0, with_fact=True, scale=None,
                  fontsize=8, **kwargs) -> Figure:
        """
        Plot the local part of the pseudopotential in q space.
        """
        nrows, ncols = len(self), len(ders)
        ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                               sharex=True, sharey=False, squeeze=False)

        fig.suptitle(_Q2VQ_LATEX)
        for i, (label, psps) in enumerate(self.items()):
            kws = dict(color=self._mkcolor(i), show=False)
            for j, der in enumerate(ders):
                #_ider = {0: 0, 1: 2}[der]
                ax = ax_mat[i,j]
                psps.plot_q2vq(ax=ax, ders=der, with_qn=with_qn, scale=scale, **kws)
                _ = "$f(q)$" if der == 0 else "$f''(q)$"
                ax.set_title(_, fontsize=fontsize) if i == 0 else ax.set_title("")
                if i != len(self) - 1: set_visible(ax, False, "xlabel")

        return fig

    @add_fig_kwargs
    def plot_ffspl(self, ecut_ffnl=None, ders=(0, 1), with_qn=0,
                   scale=None, fontsize=8, **kwargs) -> Figure:
        """
        Plot the nonlocal part of the pseudopotential in q-space.
        """
        l_select = [0, 1, 2]
        nrows, ncols = len(self), len(l_select)
        ax_mat, fig, plt = get_axarray_fig_plt(None, nrows=nrows, ncols=ncols,
                                               sharex=True, sharey=False, squeeze=False)

        fig.suptitle(f"ffnl in q-space")
        for i, (label, psps) in enumerate(self.items()):
            kws = dict(color=self._mkcolor(i), show=False)
            for j, l in enumerate(l_select):
                ax = ax_mat[i,j]
                psps.plot_ffspl(ax=ax, ders=ders, l_select=l, with_qn=with_qn, scale=scale, **kws)
                #ax.set_title(f"$ff_{nl}{der}(q)$", fontsize=fontsize) if i == 0 else ax.set_title("")
                if i != len(self) - 1: set_visible(ax, False, "xlabel")

        #fig.suptitle(r"$V_{loc} in q-space$")

        return fig

    def yield_figs(self, **kwargs):  # pragma: no cover
        """
        This function *generates* a predefined list of matplotlib figures with minimal input from the user.
        """
        verbose = kwargs.get("verbose", 0)
        kws = dict(show=False, tight_layout=True)
        yield self.plot_ffspl(**kws)
        yield self.plot_q2vq(**kws)
        yield self.plot_tcore_rspace(**kws)
        yield self.plot_tcore_qspace(**kws)

    #def yield_plotly_figs(self, **kwargs):  # pragma: no cover
    #    """
    #    This function *generates* a predefined list of plotly figures with minimal input from the user.
    #    """
    #    verbose = kwargs.get("verbose", 0)
    #    for fig in self.yield_ebands_plotly_figs(**kwargs): yield fig
    #    if verbose:
    #        for fig in self.yield_structure_plotly_figs(**kwargs): yield fig

    def write_notebook(self, nbpath=None) -> str:
        """
        Write a jupyter_ notebook to ``nbpath``. If nbpath is None, a temporay file in the current
        working directory is created. Return path to the notebook.
        """
        nbformat, nbv, nb = self.get_nbformat_nbv_nb(title=None)

        args = [(l, f.filepath) for l, f in self.items()]
        nb.cells.extend([
            #nbv.new_markdown_cell("# This is a markdown cell"),
            nbv.new_code_cell("robot = abilab.PspsRobot(*%s)\nrobot.trim_paths()\nrobot" % str(args)),
        ])

        return self._write_nb_nbpath(nb, nbpath)


class PspsReader(ETSF_Reader):
    """
    This object reads the results stored in the PSPS file produced by ABINIT.
    It provides helper functions to access the most important quantities.
    """
    def __init__(self, filepath: str):
        super().__init__(filepath)

        # Get important dimensions.
        self.usepaw, self.useylm = self.read_value("usepaw"), self.read_value("useylm")
        assert self.usepaw == 0 and self.useylm == 0
        self.ntypat = self.read_dimvalue("ntypat")
        self.lmnmax = self.read_dimvalue("lmnmax")
        self.indlmn = self.read_value("indlmn")

        self.znucl_typat = self.read_value("znucltypat")
        self.zion_typat = self.read_value("ziontypat")

    def read_coresd(self, rmax=None) -> tuple[np.ndarray, np.ndarray]:
        """
        Read the core charges and real-space derivatives for the different types of atoms.

        Args:
            rmax: Maximum radius in Bohr. If None, data on the full grid is returned.

        Returns:
            meshes: List of ntypat arrays. Each array contains the linear meshes in real space.
            coresd: List with nytpat arrays of shape [6, npts].

            (np.zeros. np.zeros) if core charge is not present

        xccc1d[ntypat, n1xccc*(1-usepaw)]

        Norm-conserving psps only
        The component xccc1d(n1xccc,1,ntypat) is the pseudo-core charge
        for each type of atom, on the radial grid. The components
        xccc1d(n1xccc,ideriv,ntypat) give the ideriv-th derivative of the
        pseudo-core charge with respect to the radial distance.
        """

        xcccrc = self.read_value("xcccrc")
        try:
            all_coresd = self.read_value("xccc1d")
        except self.Error:
            # model core may not be present!
            return self.ntypat * [np.linspace(0, 6, num=100)], self.ntypat * [np.zeros((2, 100))]

        npts = all_coresd.shape[-1]
        rmeshes, coresd = [], []
        for itypat, rc in enumerate(xcccrc):
            rvals, step = np.linspace(0, rc, num=npts, retstep=True)
            ir_stop = -1
            if rmax is not None:
                # Truncate the mesh
                ir_stop = min(int(rmax / step), npts) + 1
                #print(rmax, step, ir_stop, npts)

            rmeshes.append(rvals[:ir_stop])
            coresd.append(all_coresd[itypat,:,:ir_stop])

        return rmeshes, coresd

    def read_tcorespl(self) -> tuple[np.ndarray, np.ndarray]:
        """
        Returns:
            qmesh: Linear q-mesh in q-space.
            tcorespl: array of shape [ntypat, 2, mqgrid_vl] with the pseudo core density in reciprocal space on a regular grid.
            Only if pseudo has_tcore
        """
        return self.read_value("qgrid_vl"), self.read_value("nc_tcorespl")

    def read_vlspl(self) -> tuple[np.ndarray, np.ndarray]:
        """
        Returns:
            qmesh: Linear q-mesh in G-space
            vlspl: numpy array of shape [ntypat, two, mqgrid_vl] with the local part of each type of psp in q-space
        """
        return self.read_value("qgrid_vl"), self.read_value("vlspl")

    def read_projectors(self) -> tuple[VnlProjector, ...]:
        """
        ffspl[ntypat, lnmax, 2, mqgrid_ff]
        Gives, on the radial grid, the {n,l} non-local projectors both for NC and PAW.
        """
        # ekb(dimekb, ntypat*(1-usepaw))
        # ffspl[ntypat, lnmax, 2, mqgrid_ff]
        ekb = self.read_value("ekb")
        qgrid_ff = self.read_value("qgrid_ff")
        ffspl = self.read_value("ffspl")

        projs = self.ntypat * [None]
        for itypat in range(self.ntypat):
            projs_type = []
            ln_list = self.get_lnlist_for_type(itypat)
            for i, ln in enumerate(ln_list):
                p = VnlProjector(itypat, ln, ekb[itypat, i], qgrid_ff, ffspl[itypat, i])
                projs_type.append(p)
            projs[itypat] = projs_type

        return projs

    def get_lnlist_for_type(self, itypat: int) -> list:
        """
        Return a list of (l, n) indices for this atom type.
        """
        # indlmn(6,lmn_size,ntypat) = array giving l,m,n,lm,ln,s for i=lmn
        indlmn_type = self.indlmn[itypat,:,:]

        iln0 = 0; ln_list = []
        for ilmn in range(self.lmnmax):
            iln = indlmn_type[ilmn, 4]
            if iln > iln0:
                iln0 = iln
                l = indlmn_type[ilmn, 0]  # l
                n = indlmn_type[ilmn, 2]  # n
                ln_list.append((l, n))

        return ln_list


class VnlProjector:
    """
    Data and parameters associated to a non-local projector.
    """

    def __init__(self, itypat, ln, ekb, qmesh, data):
        """
        Args:
            itypat: Type atom index (C index >= 0)
            ln: Tuple with l and n.
            ekb: KB energy in Hartree.
            qmesh: Mesh of q-points.
            data: numpy array [2, nqpt]
        """
        self.ln = ln
        self.l, self.n, self.ekb = ln[0], ln[1], ekb
        self.qmesh, self.data = qmesh, data

        assert len(self.qmesh) == len(self.data[0])

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self, verbose: int = 0) -> str:
        """String representation."""
        lines = []; app = lines.append
        app(f"Projector: l: {self.l}, n: {self.n}, ekb: {self.ekb} (Ha)")
        return "\n".join(lines)

    @property
    def ecuts(self) -> np.ndarray:
        """List of cutoff energies in Ha corresponding to self.qmesh."""
        return 2 * (np.pi * self.qmesh)**2

    @property
    def sign_sqrtekb(self) -> np.ndarray:
        return np.sign(self.ekb) * np.sqrt(np.abs(self.ekb))
